#!/usr/bin/env python
"""
Convert prepared datasets to ptychodus format

Example usage:

convert-to-ptychodus \
    --diffraction-input "data/ptychoshelves/velo_19c2_Jun_IC_fly145/data_roi0_dp.hdf5" \
    --diffraction-output "data/velo_19c2_Jun_IC_fly145_diffraction.h5" \
    --product-input "data/ptychoshelves/velo_19c2_Jun_IC_fly145/Niter100.mat" \
    --product-output "data/velo_19c2_Jun_IC_fly145_product.h5" \
    --product-name "velo_19c2_Jun_IC_fly145"
"""

from collections.abc import Sequence
from pathlib import Path
import argparse
import json
import logging
import sys

from ptychodus.model import ModelCore
import ptychodus

logger = logging.getLogger(__name__)


def main() -> int:
    prog = Path(__file__).stem.lower()
    parser = argparse.ArgumentParser(
        prog=prog, description=f'{prog} repackages prepared datasets into ptychodus formats.'
    )
    parser.add_argument(
        '--diffraction-input-type',
        default='fold_slice',
        help='Diffraction input file type.',
    )
    parser.add_argument(
        '--diffraction-input',
        metavar='DIFFRACTION_INPUT_FILE',
        type=argparse.FileType('r'),
        help='Path to the diffraction input file.',
    )
    parser.add_argument(
        '--diffraction-output',
        metavar='DIFFRACTION_OUTPUT_FILE',
        type=argparse.FileType('w'),
        help='Path to the diffraction output file.',
    )
    parser.add_argument(
        '--list-plugins',
        action='store_true',
        help='List available file reader plugins, then exit.',
    )
    parser.add_argument(
        '--log-level',
        type=int,
        default=logging.WARNING,
        help='Set Python logging level.',
    )
    parser.add_argument(
        '--override-object-type',
        metavar='OBJECT_FILE_TYPE',
        default='fold_slice',
        help='Override object file type.',
    )
    parser.add_argument(
        '--override-object',
        metavar='OBJECT_FILE',
        type=argparse.FileType('r'),
        help='Path to the object file.',
    )
    parser.add_argument(
        '--override-probe-type',
        metavar='PROBE_FILE_TYPE',
        default='fold_slice',
        help='Override probe file type.',
    )
    parser.add_argument(
        '--override-probe',
        metavar='PROBE_FILE',
        type=argparse.FileType('r'),
        help='Path to the probe file.',
    )
    parser.add_argument(
        '--override-probe-positions-type',
        metavar='PROBE_POSITIONS_FILE_TYPE',
        default='fold_slice',
        help='Override probe positions file type.',
    )
    parser.add_argument(
        '--override-probe-positions',
        metavar='PROBE_POSITIONS_FILE',
        type=argparse.FileType('r'),
        help='Path to the probe positions file.',
    )
    parser.add_argument(
        '--product-input-type',
        default='fold_slice',
        help='Product input file type.',
    )
    parser.add_argument(
        '--product-input',
        metavar='PRODUCT_INPUT_FILE',
        type=argparse.FileType('r'),
        help='Path to the product input file.',
    )
    parser.add_argument(
        '--product-name',
        help='Data product name',
    )
    parser.add_argument(
        '--product-output',
        metavar='PRODUCT_OUTPUT_FILE',
        type=argparse.FileType('w'),
        help='Path to the product output file.',
    )
    parser.add_argument(
        '-s',
        '--settings',
        metavar='SETTINGS_FILE',
        type=argparse.FileType('r'),
        help='Path to the settings file.',
    )
    parser.add_argument(
        '-v',
        '--version',
        action='version',
        version=ptychodus.VERSION_STRING,
    )

    args = parser.parse_args()
    settings_file = Path(args.settings.name) if args.settings else None

    with ModelCore(settings_file, log_level=args.log_level) as model:
        if args.list_plugins:
            plugins: dict[str, Sequence[str]] = dict()
            registry = model.plugin_registry
            plugins['diffraction_readers'] = (
                registry.diffraction_file_readers.stringify_plugin_names()
            )
            plugins['product_readers'] = registry.product_file_readers.stringify_plugin_names()
            plugins['probe_position_readers'] = (
                registry.probe_position_file_readers.stringify_plugin_names()
            )
            plugins['probe_readers'] = registry.probe_file_readers.stringify_plugin_names()
            plugins['object_readers'] = registry.object_file_readers.stringify_plugin_names()

            print(json.dumps(plugins, indent=4))

            return 0

        missing_args: list[str] = []

        if args.diffraction_input is None:
            missing_args.append('--diffraction-input')

        if args.diffraction_output is None:
            missing_args.append('--diffraction-output')

        if args.product_input is None:
            missing_args.append('--product-input')

        if args.product_output is None:
            missing_args.append('--product-output')

        if missing_args:
            missing_args_str = ', '.join(missing_args)
            parser.error(f'the following arguments are required: {missing_args_str}')

        workflow_diffraction_api = model.workflow_api.load_diffraction_data(
            Path(args.diffraction_input.name),
            file_type=args.diffraction_input_type,
            process_patterns=False,
            block=True,
        )
        workflow_diffraction_api.save_assembled_data(args.diffraction_output.name)
        logger.info(f'Wrote diffraction data to {args.diffraction_output.name}')

        product_api = model.workflow_api.load_product(
            Path(args.product_input.name),
            file_type=args.product_input_type,
        )

        if args.product_name is not None:
            product_api.rename_product(args.product_name)

        if args.override_object is not None:
            product_api.load_object(
                Path(args.override_object.name), file_type=args.override_object_type
            )

        if args.override_probe is not None:
            product_api.load_probe(
                Path(args.override_probe.name), file_type=args.override_probe_type
            )

        if args.override_probe_positions is not None:
            product_api.load_probe_positions(
                Path(args.override_probe_positions.name),
                file_type=args.override_probe_positions_type,
            )

        product_api.save_product(Path(args.product_output.name), file_type='HDF5')
        logger.info(f'Wrote product data to {args.product_output.name}')

    return 0


if __name__ == '__main__':
    sys.exit(main())



"""Common type aliases, physical constants, and utility functions used throughout the API."""

from typing import Any, Final, TypeAlias, overload

import numpy
import numpy.typing

# Type Aliases
IntegerArrayType: TypeAlias = numpy.typing.NDArray[numpy.integer[Any]]
RealArrayType: TypeAlias = numpy.typing.NDArray[numpy.floating[Any]]
ComplexArrayType: TypeAlias = numpy.typing.NDArray[numpy.complexfloating[Any, Any]]
InexactArrayType: TypeAlias = numpy.typing.NDArray[numpy.inexact[Any]]
NumberArrayType: TypeAlias = numpy.typing.NDArray[numpy.number]

# Mathematical Constants
BYTES_PER_MEGABYTE: Final[int] = 1000 * 1000
TWO_PI: Final[float] = 2.0 * numpy.pi
TWO_PI_J: Final[complex] = 2.0j * numpy.pi

# Physical Constants
# Source: https://physics.nist.gov/cuu/Constants/index.html
ELECTRON_VOLT_J: Final[float] = 1.602176634e-19
LIGHT_SPEED_M_PER_S: Final[float] = 299792458
PLANCK_CONSTANT_J_PER_HZ: Final[float] = 6.62607015e-34


@overload
def lerp(lower: float, upper: float, frac: float) -> float: ...
@overload
def lerp(lower: complex, upper: complex, frac: float) -> complex: ...
@overload
def lerp(lower: RealArrayType, upper: RealArrayType, frac: float) -> RealArrayType: ...
@overload
def lerp(lower: RealArrayType, upper: RealArrayType, frac: RealArrayType) -> RealArrayType: ...
@overload
def lerp(lower: float, upper: float, frac: RealArrayType) -> RealArrayType: ...


def lerp(
    lower: InexactArrayType | complex,
    upper: InexactArrayType | complex,
    frac: RealArrayType | float,
) -> InexactArrayType | complex:
    """Linearly interpolate between *lower* and *upper* by fraction *frac* in [0, 1]."""
    return (1.0 - frac) * lower + frac * upper



from collections.abc import Mapping
from pathlib import Path
import logging
import re

import h5py
import numpy

from ptychodus.api.geometry import ImageExtent
from ptychodus.api.diffraction import (
    DiffractionDataset,
    DiffractionFileReader,
    DiffractionMetadata,
    DiffractionArray,
    SimpleDiffractionDataset,
)
from ptychodus.api.plugins import PluginRegistry
from ptychodus.api.tree import SimpleTreeNode

from .h5_diffraction_file import H5DiffractionPatternArray

logger = logging.getLogger(__name__)


class APS2IDDiffractionFileReader(DiffractionFileReader):
    def _get_file_series(self, file_path: Path) -> tuple[Mapping[int, Path], str]:
        file_path_dict: dict[int, Path] = dict()

        digits = re.findall(r'\d+', file_path.stem)
        longest_digits = max(digits, key=len)
        file_pattern = file_path.name.replace(longest_digits, f'(\\d{{{len(longest_digits)}}})')

        for fp in file_path.parent.iterdir():
            z = re.match(file_pattern, fp.name)

            if z:
                index = int(z.group(1))
                file_path_dict[index] = fp

        return file_path_dict, file_pattern

    def read(self, file_path: Path) -> DiffractionDataset:
        file_path_mapping, file_pattern = self._get_file_series(file_path)
        data_path = '/entry/data/data'

        with h5py.File(file_path, 'r') as h5_file:
            h5data = h5_file[data_path]

            if isinstance(h5data, h5py.Dataset):
                num_patterns_per_array, detector_height, detector_width = h5data.shape
                metadata = DiffractionMetadata(
                    num_patterns_per_array=[num_patterns_per_array] * len(file_path_mapping),
                    pattern_dtype=h5data.dtype,
                    detector_extent=ImageExtent(detector_width, detector_height),
                    file_path=file_path.parent / file_pattern,
                )
                contents_tree = SimpleTreeNode.create_root(['Name', 'Type', 'Details'])
                array_list: list[DiffractionArray] = list()

                for idx, fp in sorted(file_path_mapping.items()):
                    indexes = numpy.arange(num_patterns_per_array) + idx * num_patterns_per_array
                    array = H5DiffractionPatternArray(fp.stem, indexes, fp, data_path)
                    contents_tree.create_child([array.get_label(), 'HDF5', str(idx)])
                    array_list.append(array)

                return SimpleDiffractionDataset(metadata, contents_tree, array_list)
            else:
                raise ValueError(f'Expected dataset at "{data_path}"; found {type(h5data)}.')


def register_plugins(registry: PluginRegistry) -> None:
    registry.diffraction_file_readers.register_plugin(
        APS2IDDiffractionFileReader(),
        simple_name='APS_2IDD',
        display_name='APS 2-ID-D Microprobe Files (*.h5 *.hdf5)',
    )
    registry.diffraction_file_readers.register_plugin(
        APS2IDDiffractionFileReader(),
        simple_name='APS_2IDE',
        display_name='APS 2-ID-E Microprobe Files (*.h5 *.hdf5)',
    )
    registry.diffraction_file_readers.register_plugin(
        APS2IDDiffractionFileReader(),
        simple_name='APS_BNP',
        display_name='APS 2-ID-D Bionanoprobe Files (*.h5 *.hdf5)',
    )



from pathlib import Path
from typing import Final
import logging

import h5py

from ptychodus.api.plugins import PluginRegistry
from ptychodus.api.probe_positions import (
    ProbePositionSequence,
    ProbePositionFileReader,
    ProbePosition,
)

logger = logging.getLogger(__name__)


class NSLS2Style1PositionFileReader(ProbePositionFileReader):
    ONE_MICRON_M: Final[float] = 1.0e-6

    def read(self, file_path: Path) -> ProbePositionSequence:
        point_list: list[ProbePosition] = list()

        with h5py.File(file_path, 'r') as h5_file:
            h5_positions = h5_file['/scan/scan_positions']

            for idx, row in enumerate(h5_positions[()].T):
                point = ProbePosition(
                    idx,
                    row[0] * self.ONE_MICRON_M,
                    row[1] * self.ONE_MICRON_M,
                )
                point_list.append(point)

        return ProbePositionSequence(point_list)


class NSLS2Style2PositionFileReader(ProbePositionFileReader):
    ONE_MICRON_M: Final[float] = 1.0e-6

    def read(self, file_path: Path) -> ProbePositionSequence:
        point_list: list[ProbePosition] = list()

        with h5py.File(file_path, 'r') as h5_file:
            h5_positions = h5_file['/points']

            for idx, row in enumerate(h5_positions[()].T):
                point = ProbePosition(
                    idx,
                    row[0] * self.ONE_MICRON_M,
                    row[1] * self.ONE_MICRON_M,
                )
                point_list.append(point)

        return ProbePositionSequence(point_list)


def register_plugins(registry: PluginRegistry) -> None:
    registry.probe_position_file_readers.register_plugin(
        NSLS2Style1PositionFileReader(),
        simple_name='NSLS_II_1',
        display_name='NSLS-II Style 1 Files (*.h5 *.hdf5)',
    )
    registry.probe_position_file_readers.register_plugin(
        NSLS2Style2PositionFileReader(),
        simple_name='NSLS_II_2',
        display_name='NSLS-II Style 2 Files (*.h5 *.hdf5)',
    )



from pathlib import Path
from typing import Final
import logging

import h5py

from ptychodus.api.plugins import PluginRegistry
from ptychodus.api.probe_positions import (
    ProbePositionSequence,
    ProbePositionFileReader,
    ProbePosition,
    ProbePositionParseError,
)

logger = logging.getLogger(__name__)


class NanoMAXPositionFileReader(ProbePositionFileReader):
    ONE_MICRON_M: Final[float] = 1.0e-6

    def read(self, file_path: Path) -> ProbePositionSequence:
        point_list: list[ProbePosition] = list()

        with h5py.File(file_path, 'r') as h5_file:
            try:
                position_x = h5_file['/entry/measurement/pseudo/x'][()]
                position_y = h5_file['/entry/measurement/pseudo/y'][()]
            except KeyError:
                logger.exception('Unable to load scan.')
            else:
                if position_x.shape == position_y.shape:
                    logger.debug(f'Coordinate arrays have shape {position_x.shape}.')
                else:
                    raise ProbePositionParseError('Coordinate array shape mismatch!')

                for idx, (x, y) in enumerate(zip(position_x, position_y)):
                    point = ProbePosition(
                        idx,
                        x * self.ONE_MICRON_M,
                        y * self.ONE_MICRON_M,
                    )
                    point_list.append(point)

        return ProbePositionSequence(point_list)


def register_plugins(registry: PluginRegistry) -> None:
    registry.probe_position_file_readers.register_plugin(
        NanoMAXPositionFileReader(),
        simple_name='MAX_IV_NanoMAX',
        display_name='MAX IV NanoMAX Diffraction Endstation Files (*.h5 *.hdf5)',
    )



from collections.abc import Sequence
from dataclasses import dataclass
from enum import Enum, auto
from pathlib import Path
from typing import overload
import logging

from ptychodus.api.observer import ObservableSequence

from .settings import AutomationSettings

logger = logging.getLogger(__name__)


class AutomationDatasetState(Enum):
    FOUND = auto()
    WAITING = auto()
    RUNNING = auto()
    COMPLETE = auto()


@dataclass(frozen=True)
class AutomationDataset:
    file_path: Path
    state: AutomationDatasetState
    label: str


class AutomationDatasetRepository(ObservableSequence[AutomationDataset]):
    def __init__(self, settings: AutomationSettings) -> None:
        super().__init__()
        self._settings = settings
        self._dataset_list: list[AutomationDataset] = list()

    @overload
    def __getitem__(self, index: int) -> AutomationDataset: ...

    @overload
    def __getitem__(self, index: slice) -> Sequence[AutomationDataset]: ...

    def __getitem__(self, index: int | slice) -> AutomationDataset | Sequence[AutomationDataset]:
        return self._dataset_list[index]

    def __len__(self) -> int:
        return len(self._dataset_list)

    def upsert(self, file_path: Path, state: AutomationDatasetState) -> None:
        try:
            index = next(
                idx for idx, item in enumerate(self._dataset_list) if item.file_path == file_path
            )
        except StopIteration:
            index = len(self._dataset_list)
            label = str(file_path.relative_to(self._settings.data_directory.get_value()))
            dataset = AutomationDataset(file_path, state, label)
            self._dataset_list.append(dataset)
            self.notify_observers_item_inserted(index, dataset)
        else:
            old_dataset = self._dataset_list[index]
            new_dataset = AutomationDataset(file_path, state, old_dataset.label)
            logger.debug(f'{file_path}: {old_dataset.state} -> {new_dataset.state}')
            self._dataset_list[index] = new_dataset
            self.notify_observers_item_changed(index, new_dataset)

    def clear(self) -> None:
        while True:
            try:
                dataset = self._dataset_list.pop()
            except IndexError:
                break
            else:
                self.notify_observers_item_removed(len(self._dataset_list), dataset)


class UpdateAutomationDatasetRepository:
    def __init__(
        self,
        repository: AutomationDatasetRepository,
        file_path: Path,
        state: AutomationDatasetState,
    ) -> None:
        self._repository = repository
        self._file_path = file_path
        self._state = state

    def __call__(self) -> None:
        self._repository.upsert(self._file_path, self._state)



import numpy

from ptychodus.api.probe_positions import ScanBoundingBox, ProbePosition


class ScanBoundingBoxBuilder:
    def __init__(self) -> None:
        self._xmin_m = +numpy.inf
        self._xmax_m = -numpy.inf
        self._ymin_m = +numpy.inf
        self._ymax_m = -numpy.inf

    def hull(self, point: ProbePosition) -> None:
        if point.coordinate_x_m < self._xmin_m:
            self._xmin_m = point.coordinate_x_m

        if self._xmax_m < point.coordinate_x_m:
            self._xmax_m = point.coordinate_x_m

        if point.coordinate_y_m < self._ymin_m:
            self._ymin_m = point.coordinate_y_m

        if self._ymax_m < point.coordinate_y_m:
            self._ymax_m = point.coordinate_y_m

    def get_bounding_box(self) -> ScanBoundingBox | None:
        is_empty_x = self._xmax_m < self._xmin_m
        is_empty_y = self._ymax_m < self._ymin_m

        if is_empty_x or is_empty_y:
            return None

        return ScanBoundingBox(
            minimum_x_m=self._xmin_m,
            maximum_x_m=self._xmax_m,
            minimum_y_m=self._ymin_m,
            maximum_y_m=self._ymax_m,
        )



from collections.abc import Iterator, Mapping, Sequence
from pathlib import Path
from typing import Any
import logging

from ptychodus.api.plugins import PluginChooser
from ptychodus.api.product import Product, ProductFileReader, ProductFileWriter

from .item import ProductRepositoryItem
from .item_factory import ProductRepositoryItemFactory
from .object.builder_factory import ObjectBuilderFactory
from .object.settings import ObjectSettings
from .object_repository import ObjectRepository
from .probe.builder_factory import ProbeBuilderFactory
from .probe.settings import ProbeSettings
from .probe_repository import ProbeRepository
from .repository import ProductRepository
from .probe_positions.builder_factory import ProbePositionsBuilderFactory
from .probe_positions.settings import ProbePositionsSettings
from .scan_repository import ProbePositionsRepository
from .settings import ProductSettings

logger = logging.getLogger(__name__)


class PositionsStreamingContext:
    def __init__(self) -> None:
        self._positions_x_m: list[float] = []
        self._triggers_x: list[int] = []
        self._positions_y_m: list[float] = []
        self._triggers_y: list[int] = []

    def start(self) -> None:
        self._positions_x_m.clear()
        self._triggers_x.clear()
        self._positions_y_m.clear()
        self._triggers_y.clear()

    def append_positions_x(self, values_m: Sequence[float], trigger_counts: Sequence[int]) -> None:
        self._positions_x_m.extend(values_m)
        self._triggers_x.extend(trigger_counts)

    def append_positions_y(self, values_m: Sequence[float], trigger_counts: Sequence[int]) -> None:
        self._positions_y_m.extend(values_m)
        self._triggers_y.extend(trigger_counts)

    def stop(self) -> None:
        pass  # TODO


class ProbePositionsAPI:
    def __init__(
        self,
        settings: ProbePositionsSettings,
        repository: ProbePositionsRepository,
        builder_factory: ProbePositionsBuilderFactory,
    ) -> None:
        self._settings = settings
        self._repository = repository
        self._builder_factory = builder_factory

    def create_streaming_context(self) -> PositionsStreamingContext:
        return PositionsStreamingContext()

    def builder_names(self) -> Iterator[str]:
        return iter(self._builder_factory)

    def build_probe_positions(
        self, index: int, builder_name: str, builder_parameters: Mapping[str, Any] | None = None
    ) -> None:
        try:
            item = self._repository[index]
        except IndexError:
            logger.warning(f'Failed to access item {index}!')
            return

        try:
            builder = self._builder_factory.create(builder_name)
        except KeyError:
            logger.warning(f'Failed to create builder {builder_name}!')
            return

        if builder_parameters is not None:
            for parameter_name, parameter_value in builder_parameters.items():
                try:
                    parameter = builder.parameters()[parameter_name]
                except KeyError:
                    logger.warning(
                        f'Scan builder "{builder.get_name()}" does not have parameter "{parameter_name}"!'
                    )
                else:
                    parameter.set_value(parameter_value)

        item.set_builder(builder)

    def build_probe_positions_from_settings(self, index: int) -> None:
        try:
            item = self._repository[index]
        except IndexError:
            logger.warning(f'Failed to access item {index}!')
            return

        try:
            builder = self._builder_factory.create_from_settings()
        except KeyError:
            logger.warning('Failed to create builder from settings!')
            return

        item.set_builder(builder)

    def get_open_file_filters(self) -> Iterator[str]:
        return self._builder_factory.get_open_file_filters()

    def get_open_file_filter(self) -> str:
        return self._builder_factory.get_open_file_filter()

    def open_probe_positions(
        self, index: int, file_path: Path, *, file_type: str | None = None
    ) -> None:
        builder = self._builder_factory.create_scan_from_file(
            file_path,
            self._settings.file_type.get_value() if file_type is None else file_type,
        )

        try:
            item = self._repository[index]
        except IndexError:
            logger.warning(f'Failed to open scan {index}!')
        else:
            item.set_builder(builder)

    def copy_scan(self, source_index: int, destination_index: int) -> None:
        logger.debug(f'Copying {source_index} -> {destination_index}')

        try:
            source_item = self._repository[source_index]
        except IndexError:
            logger.warning(f'Failed to access source scan {source_index} for copying!')
            return

        try:
            destination_item = self._repository[destination_index]
        except IndexError:
            logger.warning(f'Failed to access destination scan {destination_index} for copying!')
            return

        destination_item.assign_item(source_item)

    def get_save_file_filters(self) -> Iterator[str]:
        return self._builder_factory.get_save_file_filters()

    def get_save_file_filter(self) -> str:
        return self._builder_factory.get_save_file_filter()

    def save_scan(self, index: int, file_path: Path, file_type: str) -> None:
        try:
            item = self._repository[index]
        except IndexError:
            logger.warning(f'Failed to save scan {index}!')
        else:
            self._builder_factory.save_scan(file_path, file_type, item.get_probe_positions())


class ProbeAPI:
    def __init__(
        self,
        settings: ProbeSettings,
        repository: ProbeRepository,
        builder_factory: ProbeBuilderFactory,
    ) -> None:
        self._settings = settings
        self._repository = repository
        self._builder_factory = builder_factory

    def builder_names(self) -> Iterator[str]:
        return iter(self._builder_factory)

    def build_probe(
        self, index: int, builder_name: str, builder_parameters: Mapping[str, Any] | None = None
    ) -> None:
        try:
            item = self._repository[index]
        except IndexError:
            logger.warning(f'Failed to access item {index}!')
            return

        try:
            builder = self._builder_factory.create(builder_name)
        except KeyError:
            logger.warning(f'Failed to create builder {builder_name}!')
            return

        if builder_parameters is not None:
            for parameter_name, parameter_value in builder_parameters.items():
                try:
                    parameter = builder.parameters()[parameter_name]
                except KeyError:
                    logger.warning(
                        f'Probe builder "{builder.get_name()}" does not have'
                        f' parameter "{parameter_name}"!'
                    )
                else:
                    parameter.set_value(parameter_value)

        item.set_builder(builder)

    def build_probe_from_settings(self, index: int) -> None:
        try:
            item = self._repository[index]
        except IndexError:
            logger.warning(f'Failed to access item {index}!')
            return

        try:
            builder = self._builder_factory.create_from_settings()
        except KeyError:
            logger.warning('Failed to create builder from settings!')
            return

        item.set_builder(builder)

    def get_open_file_filters(self) -> Iterator[str]:
        return self._builder_factory.get_open_file_filters()

    def get_open_file_filter(self) -> str:
        return self._builder_factory.get_open_file_filter()

    def open_probe(self, index: int, file_path: Path, *, file_type: str | None = None) -> None:
        builder = self._builder_factory.create_probe_from_file(
            file_path,
            self._settings.file_type.get_value() if file_type is None else file_type,
        )

        try:
            item = self._repository[index]
        except IndexError:
            logger.warning(f'Failed to open probe {index}!')
        else:
            item.set_builder(builder)

    def copy_probe(self, source_index: int, destination_index: int) -> None:
        logger.debug(f'Copying {source_index} -> {destination_index}')

        try:
            source_item = self._repository[source_index]
        except IndexError:
            logger.warning(f'Failed to access source probe {source_index} for copying!')
            return

        try:
            destination_item = self._repository[destination_index]
        except IndexError:
            logger.warning(f'Failed to access destination probe {destination_index} for copying!')
            return

        destination_item.assign_item(source_item)

    def get_save_file_filters(self) -> Iterator[str]:
        return self._builder_factory.get_save_file_filters()

    def get_save_file_filter(self) -> str:
        return self._builder_factory.get_save_file_filter()

    def save_probe(self, index: int, file_path: Path, file_type: str) -> None:
        try:
            item = self._repository[index]
        except IndexError:
            logger.warning(f'Failed to save probe {index}!')
        else:
            self._builder_factory.save_probe(file_path, file_type, item.get_probes())


class ObjectAPI:
    def __init__(
        self,
        settings: ObjectSettings,
        repository: ObjectRepository,
        builder_factory: ObjectBuilderFactory,
    ) -> None:
        self._settings = settings
        self._repository = repository
        self._builder_factory = builder_factory

    def builder_names(self) -> Iterator[str]:
        return iter(self._builder_factory)

    def build_object(
        self, index: int, builder_name: str, builder_parameters: Mapping[str, Any] | None = None
    ) -> None:
        try:
            item = self._repository[index]
        except IndexError:
            logger.warning(f'Failed to access item {index}!')
            return

        try:
            builder = self._builder_factory.create(builder_name)
        except KeyError:
            logger.warning(f'Failed to create builder {builder_name}!')
            return

        if builder_parameters is not None:
            for parameter_name, parameter_value in builder_parameters.items():
                try:
                    parameter = builder.parameters()[parameter_name]
                except KeyError:
                    logger.warning(
                        f'Object builder "{builder.get_name()}" does not have'
                        f' parameter "{parameter_name}"!'
                    )
                else:
                    parameter.set_value(parameter_value)

        item.set_builder(builder)

    def build_object_from_settings(self, index: int) -> None:
        try:
            item = self._repository[index]
        except IndexError:
            logger.warning(f'Failed to access item {index}!')
            return

        try:
            builder = self._builder_factory.create_from_settings()
        except KeyError:
            logger.warning('Failed to create builder from settings!')
            return

        item.set_builder(builder)

    def get_open_file_filters(self) -> Iterator[str]:
        return self._builder_factory.get_open_file_filters()

    def get_open_file_filter(self) -> str:
        return self._builder_factory.get_open_file_filter()

    def open_object(self, index: int, file_path: Path, *, file_type: str | None = None) -> None:
        builder = self._builder_factory.create_object_from_file(
            file_path,
            self._settings.file_type.get_value() if file_type is None else file_type,
        )

        try:
            item = self._repository[index]
        except IndexError:
            logger.warning(f'Failed to open object {index}!')
        else:
            item.set_builder(builder)

    def copy_object(self, source_index: int, destination_index: int) -> None:
        logger.debug(f'Copying {source_index} -> {destination_index}')

        try:
            source_item = self._repository[source_index]
        except IndexError:
            logger.warning(f'Failed to access source object {source_index} for copying!')
            return

        try:
            destination_item = self._repository[destination_index]
        except IndexError:
            logger.warning(f'Failed to access destination object {destination_index} for copying!')
            return

        destination_item.assign_item(source_item)

    def get_save_file_filters(self) -> Iterator[str]:
        return self._builder_factory.get_save_file_filters()

    def get_save_file_filter(self) -> str:
        return self._builder_factory.get_save_file_filter()

    def save_object(self, index: int, file_path: Path, file_type: str) -> None:
        try:
            item = self._repository[index]
        except IndexError:
            logger.warning(f'Failed to save object {index}!')
        else:
            self._builder_factory.save_object(file_path, file_type, item.get_object())


class ProductAPI:
    def __init__(
        self,
        settings: ProductSettings,
        repository: ProductRepository,
        item_factory: ProductRepositoryItemFactory,
        file_reader_chooser: PluginChooser[ProductFileReader],
        file_writer_chooser: PluginChooser[ProductFileWriter],
    ) -> None:
        self._settings = settings
        self._repository = repository
        self._item_factory = item_factory
        self._file_reader_chooser = file_reader_chooser
        self._file_writer_chooser = file_writer_chooser

    def insert_new_product(
        self,
        name: str = 'Unnamed',
        *,
        comments: str = '',
        detector_distance_m: float | None = None,
        probe_energy_eV: float | None = None,  # noqa: N803
        probe_photon_count: float | None = None,
        exposure_time_s: float | None = None,
        mass_attenuation_m2_kg: float | None = None,
        tomography_angle_deg: float | None = None,
    ) -> int:
        item = self._item_factory.create_from_values(
            name=name,
            comments=comments,
            detector_distance_m=detector_distance_m,
            probe_energy_eV=probe_energy_eV,
            probe_photon_count=probe_photon_count,
            exposure_time_s=exposure_time_s,
            mass_attenuation_m2_kg=mass_attenuation_m2_kg,
            tomography_angle_deg=tomography_angle_deg,
        )
        return self._repository.insert_product(item)

    def insert_product(self, product: Product) -> int:
        item = self._item_factory.create_from_product(product)
        return self._repository.insert_product(item)

    def insert_product_from_settings(self) -> int:
        item = self._item_factory.create_from_settings()
        return self._repository.insert_product(item)

    def get_item(self, product_index: int) -> ProductRepositoryItem:
        return self._repository[product_index]

    def get_open_file_filters(self) -> Iterator[str]:
        for plugin in self._file_reader_chooser:
            yield plugin.display_name

    def get_open_file_filter(self) -> str:
        return self._file_reader_chooser.get_current_plugin().display_name

    def open_product(self, file_path: Path, *, file_type: str | None = None) -> int:
        if file_path.is_file():
            if file_type is not None:
                self._file_reader_chooser.set_current_plugin(file_type)

            file_type = self._file_reader_chooser.get_current_plugin().simple_name
            logger.debug(f'Reading "{file_path}" as "{file_type}"')
            file_reader = self._file_reader_chooser.get_current_plugin().strategy

            try:
                product = file_reader.read(file_path)
            except Exception as exc:
                raise RuntimeError(f'Failed to read "{file_path}"') from exc
            else:
                return self.insert_product(product)
        else:
            logger.warning(f'Refusing to create product with invalid file path "{file_path}"')

        return -1

    def rename_product(self, product_index: int, new_name: str) -> None:
        try:
            item = self._repository[product_index]
        except IndexError:
            logger.warning(f'Failed to access product {product_index} for renaming!')
            return

        item.set_name(new_name)

    def get_save_file_filters(self) -> Iterator[str]:
        for plugin in self._file_writer_chooser:
            yield plugin.display_name

    def get_save_file_filter(self) -> str:
        return self._file_writer_chooser.get_current_plugin().display_name

    def save_product(self, index: int, file_path: Path, *, file_type: str | None = None) -> None:
        try:
            item = self._repository[index]
        except IndexError:
            logger.warning(f'Failed to save product {index}!')
            return

        if file_type is not None:
            self._file_writer_chooser.set_current_plugin(file_type)

        file_type = self._file_writer_chooser.get_current_plugin().simple_name
        logger.debug(f'Writing "{file_path}" as "{file_type}"')
        writer = self._file_writer_chooser.get_current_plugin().strategy
        writer.write(file_path, item.get_product())



from __future__ import annotations
from typing import Any, overload

import numpy

from PyQt5.QtCore import Qt, QAbstractItemModel, QModelIndex, QObject
from PyQt5.QtGui import QBrush

from ptychodus.api.common import BYTES_PER_MEGABYTE
from ptychodus.api.probe import Probe

from ...model.product import ProbeAPI, ProbeRepository
from ...model.product.probe import ProbeRepositoryItem


class ProbeTreeNode:
    def __init__(self, parent: ProbeTreeNode | None = None) -> None:
        self.parent = parent
        self.children: list[ProbeTreeNode] = list()

    def insert_node(self, index: int = -1) -> ProbeTreeNode:
        node = ProbeTreeNode(self)
        self.children.insert(index, node)
        return node

    def remove_node(self, index: int = -1) -> ProbeTreeNode:
        return self.children.pop(index)

    def row(self) -> int:
        return 0 if self.parent is None else self.parent.children.index(self)


def calc_relative_power_percent(probe: Probe, imode: int) -> int:
    try:
        relative_power = probe.get_incoherent_mode_relative_power(imode)
    except IndexError:
        return -1

    if numpy.isfinite(relative_power):
        return int(100.0 * relative_power)

    return -1


def calc_coherent_percent(probe: Probe) -> int:
    coherence = probe.get_coherence()
    return int(100.0 * coherence) if numpy.isfinite(coherence) else -1


class ProbeTreeModel(QAbstractItemModel):
    def __init__(
        self,
        repository: ProbeRepository,
        api: ProbeAPI,
        editable_item_brush: QBrush,
        parent: QObject | None = None,
    ) -> None:
        super().__init__(parent)
        self._repository = repository
        self._api = api
        self._editable_item_brush = editable_item_brush
        self._tree_root = ProbeTreeNode()
        self._header = [
            'Name',
            'Relative Power',
            'Builder',
            'Data Type',
            'Width [px]',
            'Height [px]',
            'Pixel Width\n[nm]',
            'Pixel Height\n[nm]',
            'Size [MB]',
        ]

        for index, item in enumerate(repository):
            self.insert_item(index, item)

    @staticmethod
    def _append_modes(node: ProbeTreeNode, item: ProbeRepositoryItem) -> None:
        probe = item.get_probes()

        for layer in range(probe.num_incoherent_modes):
            node.insert_node()

    def insert_item(self, index: int, item: ProbeRepositoryItem) -> None:
        self.beginInsertRows(QModelIndex(), index, index)
        ProbeTreeModel._append_modes(self._tree_root.insert_node(index), item)
        self.endInsertRows()

    def update_item(self, index: int, item: ProbeRepositoryItem) -> None:
        top_left = self.index(index, 0)
        bottom_right = self.index(index, len(self._header))
        self.dataChanged.emit(top_left, bottom_right)

        node = self._tree_root.children[index]
        num_modes_old = len(node.children)
        num_modes_new = item.get_probes().num_incoherent_modes

        if num_modes_old < num_modes_new:
            self.beginInsertRows(top_left, num_modes_old, num_modes_new)

            while len(node.children) < num_modes_new:
                node.insert_node()

            self.endInsertRows()
        elif num_modes_old > num_modes_new:
            self.beginRemoveRows(top_left, num_modes_new, num_modes_old)

            while len(node.children) > num_modes_new:
                node.remove_node()

            self.endRemoveRows()

        child_top_left = self.index(0, 0, top_left)
        child_bottom_right = self.index(num_modes_new, len(self._header), top_left)
        self.dataChanged.emit(child_top_left, child_bottom_right)

    def remove_item(self, index: int, item: ProbeRepositoryItem) -> None:
        self.beginRemoveRows(QModelIndex(), index, index)
        self._tree_root.remove_node(index)
        self.endRemoveRows()

    def headerData(  # noqa: N802
        self,
        section: int,
        orientation: Qt.Orientation,
        role: int = Qt.ItemDataRole.DisplayRole,
    ) -> Any:
        if orientation == Qt.Orientation.Horizontal and role == Qt.ItemDataRole.DisplayRole:
            return self._header[section]

    @overload
    def parent(self, child: QModelIndex) -> QModelIndex: ...

    @overload
    def parent(self) -> QObject: ...

    def parent(self, child: QModelIndex | None = None) -> QModelIndex | QObject:
        if child is None:
            return super().parent()
        elif child.isValid():
            node = child.internalPointer()
            parent_node = node.parent
            return (
                QModelIndex()
                if parent_node is self._tree_root
                else self.createIndex(parent_node.row(), 0, parent_node)
            )

        return QModelIndex()

    def index(self, row: int, column: int, parent: QModelIndex = QModelIndex()) -> QModelIndex:
        if self.hasIndex(row, column, parent):
            if parent.isValid():
                parent_node = parent.internalPointer()
                node = parent_node.children[row]
            else:
                node = self._tree_root.children[row]

            return self.createIndex(row, column, node)

        return QModelIndex()

    def data(self, index: QModelIndex, role: int = Qt.ItemDataRole.DisplayRole) -> Any:
        if not index.isValid():
            return None

        parent = index.parent()

        if parent.isValid():
            item = self._repository[parent.row()]

            if role == Qt.ItemDataRole.DisplayRole:
                match index.column():
                    case 0:
                        return f'Mode {index.row() + 1}'
                    case 1:
                        probe = item.get_probes().get_probe_no_opr()  # TODO OPR
                        power_percent = calc_relative_power_percent(probe, index.row())
                        return f'{power_percent}%'
            elif role == Qt.ItemDataRole.BackgroundRole:
                if index.flags() & Qt.ItemFlag.ItemIsEditable:
                    return self._editable_item_brush
            elif role == Qt.ItemDataRole.UserRole and index.column() == 1:
                probe = item.get_probes().get_probe_no_opr()  # TODO OPR
                return calc_relative_power_percent(probe, index.row())
        else:
            item = self._repository[index.row()]
            probes = item.get_probes()
            probe = probes.get_probe_no_opr()  # TODO OPR
            pixel_geometry = probe.get_pixel_geometry()

            if role == Qt.ItemDataRole.DisplayRole:
                match index.column():
                    case 0:
                        return self._repository.get_name(index.row())
                    case 1:
                        coherent_percent = calc_coherent_percent(probe)
                        return f'{coherent_percent}%'
                    case 2:
                        return item.get_builder().get_name()
                    case 3:
                        return str(probe.dtype)
                    case 4:
                        return probe.width_px
                    case 5:
                        return probe.height_px
                    case 6:
                        return f'{pixel_geometry.width_m * 1e9:.4g}'
                    case 7:
                        return f'{pixel_geometry.height_m * 1e9:.4g}'
                    case 8:
                        return f'{probes.nbytes / BYTES_PER_MEGABYTE:.2f}'
            elif role == Qt.ItemDataRole.BackgroundRole:
                if index.flags() & Qt.ItemFlag.ItemIsEditable:
                    return self._editable_item_brush
            elif role == Qt.ItemDataRole.UserRole and index.column() == 1:
                probe = item.get_probes().get_probe_no_opr()  # TODO OPR
                return calc_coherent_percent(probe)

    def flags(self, index: QModelIndex) -> Qt.ItemFlags:
        value = super().flags(index)

        if index.isValid():
            parent = index.parent()

            if not parent.isValid() and index.column() in (0, 2):
                value |= Qt.ItemFlag.ItemIsEditable

        return value

    def setData(self, index: QModelIndex, value: Any, role: int = Qt.ItemDataRole.EditRole) -> bool:  # noqa: N802
        if index.isValid() and role == Qt.ItemDataRole.EditRole:
            parent = index.parent()

            if not parent.isValid():
                if index.column() == 0:
                    self._repository.set_name(index.row(), str(value))
                    return True
                elif index.column() == 2:
                    self._api.build_probe(index.row(), str(value))
                    return True

        return False

    def rowCount(self, parent: QModelIndex = QModelIndex()) -> int:  # noqa: N802
        if parent.column() > 0:
            return 0

        node = parent.internalPointer() if parent.isValid() else self._tree_root
        return len(node.children)

    def columnCount(self, parent: QModelIndex = QModelIndex()) -> int:  # noqa: N802
        return len(self._header)



from __future__ import annotations
from collections.abc import Iterator
from pathlib import Path
import logging


from ptychodus.api.fluorescence import (
    DeconvolutionStrategy,
    ElementMap,
    FluorescenceDataset,
    FluorescenceEnhancingAlgorithm,
    FluorescenceFileReader,
    FluorescenceFileWriter,
    UpscalingStrategy,
)
from ptychodus.api.geometry import PixelGeometry
from ptychodus.api.observer import Observable, Observer
from ptychodus.api.plugins import PluginChooser
from ptychodus.api.settings import SettingsRegistry

from ..product import ProductRepository, ProductRepositoryItem
from ..visualization import VisualizationEngine
from .settings import FluorescenceSettings
from .two_step import TwoStepFluorescenceEnhancingAlgorithm
from .vspi import VSPIFluorescenceEnhancingAlgorithm

logger = logging.getLogger(__name__)


class FluorescenceEnhancer(Observable, Observer):
    def __init__(
        self,
        settings: FluorescenceSettings,
        product_repository: ProductRepository,
        two_step_enhancing_algorithm: TwoStepFluorescenceEnhancingAlgorithm,
        vspi_enhancing_algorithm: VSPIFluorescenceEnhancingAlgorithm,
        file_reader_chooser: PluginChooser[FluorescenceFileReader],
        file_writer_chooser: PluginChooser[FluorescenceFileWriter],
        reinit_observable: Observable,
    ) -> None:
        super().__init__()
        self._settings = settings
        self._product_repository = product_repository
        self.two_step_enhancing_algorithm = two_step_enhancing_algorithm
        self.vspi_enhancing_algorithm = vspi_enhancing_algorithm
        self._file_reader_chooser = file_reader_chooser
        self._file_writer_chooser = file_writer_chooser
        self._reinit_observable = reinit_observable

        self._algorithm_chooser = PluginChooser[FluorescenceEnhancingAlgorithm]()
        self._algorithm_chooser.register_plugin(
            two_step_enhancing_algorithm,
            simple_name=TwoStepFluorescenceEnhancingAlgorithm.SIMPLE_NAME,
            display_name=TwoStepFluorescenceEnhancingAlgorithm.DISPLAY_NAME,
        )
        self._algorithm_chooser.register_plugin(
            vspi_enhancing_algorithm,
            simple_name=VSPIFluorescenceEnhancingAlgorithm.SIMPLE_NAME,
            display_name=VSPIFluorescenceEnhancingAlgorithm.DISPLAY_NAME,
        )
        self._algorithm_chooser.synchronize_with_parameter(settings.algorithm)
        self._algorithm_chooser.add_observer(self)

        self._product_index = -1
        self._measured: FluorescenceDataset | None = None
        self._enhanced: FluorescenceDataset | None = None

        file_reader_chooser.synchronize_with_parameter(settings.file_type)
        file_writer_chooser.set_current_plugin(settings.file_type.get_value())
        reinit_observable.add_observer(self)

    @property
    def _product(self) -> ProductRepositoryItem:
        return self._product_repository[self._product_index]

    def set_product(self, product_index: int) -> None:
        if self._product_index != product_index:
            self._product_index = product_index
            self._enhanced = None
            self.notify_observers()

    def get_product_name(self) -> str:
        return self._product.get_name()

    def get_pixel_geometry(self) -> PixelGeometry:
        return self._product.get_geometry().get_object_plane_pixel_geometry()

    def get_open_file_filters(self) -> Iterator[str]:
        for plugin in self._file_reader_chooser:
            yield plugin.display_name

    def get_open_file_filter(self) -> str:
        return self._file_reader_chooser.get_current_plugin().display_name

    def open_measured_dataset(self, file_path: Path, file_filter: str) -> None:
        if file_path.is_file():
            self._file_reader_chooser.set_current_plugin(file_filter)
            file_type = self._file_reader_chooser.get_current_plugin().simple_name
            logger.debug(f'Reading "{file_path}" as "{file_type}"')
            file_reader = self._file_reader_chooser.get_current_plugin().strategy

            try:
                measured = file_reader.read(file_path)
            except Exception as exc:
                raise RuntimeError(f'Failed to read "{file_path}"') from exc
            else:
                self._measured = measured
                self._enhanced = None

                self._settings.file_path.set_value(file_path)

                self.notify_observers()
        else:
            logger.warning(f'Refusing to load dataset from invalid file path "{file_path}"')

    def get_num_channels(self) -> int:
        return 0 if self._measured is None else len(self._measured.element_maps)

    def get_measured_element_map(self, channel_index: int) -> ElementMap:
        if self._measured is None:
            raise ValueError('Fluorescence dataset not loaded!')

        return self._measured.element_maps[channel_index]

    def algorithms(self) -> Iterator[str]:
        for plugin in self._algorithm_chooser:
            yield plugin.display_name

    def get_algorithm(self) -> str:
        return self._algorithm_chooser.get_current_plugin().display_name

    def set_algorithm(self, name: str) -> None:
        self._algorithm_chooser.set_current_plugin(name)

    def enhance_fluorescence(self) -> None:
        if self._measured is None:
            raise ValueError('Fluorescence dataset not loaded!')
        else:
            algorithm = self._algorithm_chooser.get_current_plugin().strategy
            product = self._product.get_product()
            self._enhanced = algorithm.enhance(self._measured, product)
            self.notify_observers()

    def get_enhanced_element_map(self, channel_index: int) -> ElementMap:
        if self._enhanced is None:
            return self.get_measured_element_map(channel_index)

        return self._enhanced.element_maps[channel_index]

    def get_save_file_filters(self) -> Iterator[str]:
        for plugin in self._file_writer_chooser:
            yield plugin.display_name

    def get_save_file_filter(self) -> str:
        return self._file_writer_chooser.get_current_plugin().display_name

    def save_enhanced_dataset(self, file_path: Path, file_filter: str) -> None:
        if self._enhanced is None:
            raise ValueError('Fluorescence dataset not enhanced!')

        self._file_writer_chooser.set_current_plugin(file_filter)
        file_type = self._file_writer_chooser.get_current_plugin().simple_name
        logger.debug(f'Writing "{file_path}" as "{file_type}"')
        writer = self._file_writer_chooser.get_current_plugin().strategy
        writer.write(file_path, self._enhanced)

    def _open_fluorescence_file_from_settings(self) -> None:
        self.open_measured_dataset(
            self._settings.file_path.get_value(), self._settings.file_type.get_value()
        )

    def _update(self, observable: Observable) -> None:
        if observable is self._algorithm_chooser:
            self.notify_observers()
        elif observable is self._reinit_observable:
            self._open_fluorescence_file_from_settings()


class FluorescenceCore:
    def __init__(
        self,
        settings_registry: SettingsRegistry,
        product_repository: ProductRepository,
        upscaling_strategy_chooser: PluginChooser[UpscalingStrategy],
        deconvolution_strategy_chooser: PluginChooser[DeconvolutionStrategy],
        file_reader_chooser: PluginChooser[FluorescenceFileReader],
        file_writer_chooser: PluginChooser[FluorescenceFileWriter],
    ) -> None:
        self._settings = FluorescenceSettings(settings_registry)
        self._two_step_enhancing_algorithm = TwoStepFluorescenceEnhancingAlgorithm(
            self._settings, upscaling_strategy_chooser, deconvolution_strategy_chooser
        )
        self._vspi_enhancing_algorithm = VSPIFluorescenceEnhancingAlgorithm(self._settings)

        self.enhancer = FluorescenceEnhancer(
            self._settings,
            product_repository,
            self._two_step_enhancing_algorithm,
            self._vspi_enhancing_algorithm,
            file_reader_chooser,
            file_writer_chooser,
            settings_registry,
        )
        self.visualization_engine = VisualizationEngine(is_complex=False)

    def enhance_fluorescence(
        self, product_index: int, input_file_path: Path, output_file_path: Path
    ) -> int:
        file_type = 'XRF-Maps'

        try:
            self.enhancer.set_product(product_index)
            self.enhancer.open_measured_dataset(input_file_path, file_type)
            self.enhancer.enhance_fluorescence()
            self.enhancer.save_enhanced_dataset(output_file_path, file_type)
        except Exception as exc:
            logger.exception(exc)
            return -1

        return 0



from .core import AnalysisCore
from .diffraction import DiffractionSimulator
from .fourier import FourierAnalyzer
from .frc import FourierRingCorrelator
from .illumination import IlluminationMapper, IlluminationMap
from .interpolators import BarycentricArrayInterpolator, BarycentricArrayStitcher
from .propagator import ProbePropagator
from .xmcd import XMCDAnalyzer, XMCDResult

__all__ = [
    'AnalysisCore',
    'BarycentricArrayInterpolator',
    'BarycentricArrayStitcher',
    'DiffractionSimulator',
    'FourierAnalyzer',
    'FourierRingCorrelator',
    'IlluminationMap',
    'IlluminationMapper',
    'ProbePropagator',
    'XMCDAnalyzer',
    'XMCDResult',
]



from pathlib import Path
from typing import Final, Sequence

import h5py

from ptychodus.api.geometry import PixelGeometry
from ptychodus.api.object import Object
from ptychodus.api.plugins import PluginRegistry
from ptychodus.api.probe import ProbeSequence
from ptychodus.api.product import LossValue, Product, ProductFileReader, ProductMetadata
from ptychodus.api.probe_positions import ProbePositionSequence, ProbePosition


class NSLSIIProductFileReader(ProductFileReader):
    SIMPLE_NAME: Final[str] = 'NSLS_II_MATLAB'
    DISPLAY_NAME: Final[str] = 'NSLS-II MATLAB Files (*.mat)'
    ONE_MICRON_M: Final[float] = 1.0e-6

    def read(self, file_path: Path) -> Product:
        point_list: list[ProbePosition] = list()

        with h5py.File(file_path, 'r') as h5_file:
            detector_distance_m = float(h5_file['det_dist'][()]) * self.ONE_MICRON_M
            probe_energy_eV = 1000 * float(h5_file['energy'][()])  # noqa: N806

            metadata = ProductMetadata(
                name=file_path.stem,
                comments='',
                detector_distance_m=detector_distance_m,
                probe_energy_eV=probe_energy_eV,
                probe_photon_count=0.0,  # not included in file
                exposure_time_s=0.0,  # not included in file
                mass_attenuation_m2_kg=0.0,  # not included in file
                tomography_angle_deg=0.0,  # not included in file
            )

            pixel_width_m = h5_file['img_pixel_size_x'][()]
            pixel_height_m = h5_file['img_pixel_size_y'][()]
            pixel_geometry = PixelGeometry(width_m=pixel_width_m, height_m=pixel_height_m)
            positions_m = h5_file['pos_xy'][()].T * self.ONE_MICRON_M

            for index, _xy in enumerate(positions_m):
                point = ProbePosition(
                    index=index,
                    coordinate_x_m=_xy[1],
                    coordinate_y_m=_xy[2],
                )
                point_list.append(point)

            probe_array = h5_file['prb'][()].astype(complex)
            probes = ProbeSequence(
                array=probe_array, opr_weights=None, pixel_geometry=pixel_geometry
            )

            object_array = h5_file['obj'][()].astype(complex)
            object_ = Object(
                array=object_array,
                pixel_geometry=pixel_geometry,
                center=None,
            )
            loss: Sequence[LossValue] = list()

        return Product(
            metadata=metadata,
            probe_positions=ProbePositionSequence(point_list),
            probes=probes,
            object_=object_,
            losses=loss,
        )


def register_plugins(registry: PluginRegistry) -> None:
    registry.register_product_file_reader_with_adapters(
        NSLSIIProductFileReader(),
        simple_name=NSLSIIProductFileReader.SIMPLE_NAME,
        display_name=NSLSIIProductFileReader.DISPLAY_NAME,
    )



"""Observable typed parameters and hierarchical parameter groups for settings management."""

from __future__ import annotations
from abc import ABC, abstractmethod
from collections.abc import Iterator, Mapping, MutableSequence, Sequence
from pathlib import Path
from typing import Any, Final, Generic, TypeVar
from uuid import UUID
import logging

from .observer import Observable, Observer

__all__ = [
    'Parameter',
    'ParameterGroup',
]

logger = logging.getLogger(__name__)

T = TypeVar('T')


class Parameter(ABC, Generic[T], Observable):
    """Abstract typed parameter that notifies observers when its value changes."""

    def __init__(self, parent: Parameter[T] | None = None) -> None:
        super().__init__()
        self._parent = parent

    @abstractmethod
    def get_value(self) -> T:
        pass

    @abstractmethod
    def set_value(self, value: T, *, notify: bool = True) -> None:
        pass

    @abstractmethod
    def get_value_as_string(self) -> str:
        pass

    @abstractmethod
    def set_value_from_string(self, value: str) -> None:
        pass

    @abstractmethod
    def copy(self) -> Parameter[T]:
        pass

    def sync_value_to_parent(self) -> None:
        if self._parent is None:
            logger.warning('sync_value_to_parent: parent is None!')
        else:
            self._parent.set_value(self.get_value())

    def sync_value_from_parent(self) -> None:
        if self._parent is None:
            logger.warning('sync_value_from_parent: parent is None!')
        else:
            self.set_value(self._parent.get_value())


class ParameterBase(Parameter[T]):
    """Concrete Parameter base that stores a value and notifies on change."""

    def __init__(self, value: T, parent: Parameter[T] | None) -> None:
        super().__init__(parent)
        self._value = value

    def get_value(self) -> T:
        return self._value

    def set_value(self, value: T, *, notify: bool = True) -> None:
        if self._value != value:
            self._value = value

            if notify:
                self.notify_observers()

    def get_value_as_string(self) -> str:
        return repr(self._value)


class StringParameter(ParameterBase[str]):
    """Parameter holding a str value."""

    def __init__(self, value: str, parent: StringParameter | None) -> None:
        super().__init__(value, parent)

    def set_value_from_string(self, value: str) -> None:
        self.set_value(str(value))

    def get_value_as_string(self) -> str:
        return str(self._value)

    def copy(self) -> StringParameter:
        return StringParameter(self.get_value(), self)


class PathParameter(ParameterBase[Path]):
    """Parameter holding a Path value, with optional prefix-remapping on save."""

    def __init__(self, value: Path, parent: PathParameter | None) -> None:
        super().__init__(value, parent)

    def set_value_from_string(self, value: str) -> None:
        self.set_value(Path(value))

    def change_path_prefix(self, find_path_prefix: Path, replacement_path_prefix: Path) -> Path:
        value = self.get_value()

        try:
            relative_path = value.resolve().relative_to(find_path_prefix)
        except ValueError:
            pass
        else:
            return replacement_path_prefix / relative_path

        return value

    def get_value_as_string(self) -> str:
        return str(self._value)

    def copy(self) -> PathParameter:
        return PathParameter(self.get_value(), self)


class UUIDParameter(ParameterBase[UUID]):
    """Parameter holding a UUID value."""

    def __init__(self, value: UUID, parent: UUIDParameter | None) -> None:
        super().__init__(value, parent)

    def set_value_from_string(self, value: str) -> None:
        self.set_value(UUID(value))

    def get_value_as_string(self) -> str:
        return str(self._value)

    def copy(self) -> UUIDParameter:
        return UUIDParameter(self.get_value(), self)


class BooleanParameter(ParameterBase[bool]):
    """Parameter holding a bool value, parseable from common truth strings."""

    TRUE_VALUES: Final = ('1', 'true', 't', 'yes', 'y')

    def __init__(self, value: bool, parent: BooleanParameter | None) -> None:
        super().__init__(value, parent)

    def set_value_from_string(self, value: str) -> None:
        self.set_value(value.lower() in BooleanParameter.TRUE_VALUES)

    def copy(self) -> BooleanParameter:
        return BooleanParameter(self.get_value(), self)


class IntegerParameter(ParameterBase[int]):
    """Parameter holding an int value with optional inclusive minimum and maximum bounds."""

    def __init__(
        self,
        value: int,
        parent: IntegerParameter | None,
        *,
        minimum: int | None = None,
        maximum: int | None = None,
    ) -> None:
        super().__init__(value, parent)
        self._minimum = minimum
        self._maximum = maximum

    def get_minimum(self) -> int | None:
        return self._minimum

    def get_maximum(self) -> int | None:
        return self._maximum

    def get_value(self) -> int:
        value = super().get_value()

        if self._minimum is not None:
            value = max(self._minimum, value)

        if self._maximum is not None:
            value = min(value, self._maximum)

        return value

    def set_value_from_string(self, value: str) -> None:
        self.set_value(int(value))

    def copy(self) -> IntegerParameter:
        return IntegerParameter(
            self.get_value(), self, minimum=self.get_minimum(), maximum=self.get_maximum()
        )


class RealParameter(ParameterBase[float]):
    """Parameter holding a float value with optional inclusive minimum and maximum bounds."""

    def __init__(
        self,
        value: float,
        parent: RealParameter | None,
        *,
        minimum: float | None = None,
        maximum: float | None = None,
    ) -> None:
        super().__init__(value, parent)
        self._minimum = minimum
        self._maximum = maximum

    def get_minimum(self) -> float | None:
        return self._minimum

    def get_maximum(self) -> float | None:
        return self._maximum

    def get_value(self) -> float:
        value = super().get_value()

        if self._minimum is not None:
            value = max(self._minimum, value)

        if self._maximum is not None:
            value = min(value, self._maximum)

        return value

    def set_value(self, value: float, *, notify: bool = True) -> None:
        super().set_value(float(value), notify=notify)

    def set_value_from_string(self, value: str) -> None:
        self.set_value(float(value))

    def copy(self) -> RealParameter:
        return RealParameter(
            self.get_value(), self, minimum=self.get_minimum(), maximum=self.get_maximum()
        )


class IntegerSequenceParameter(ParameterBase[MutableSequence[int]]):
    """Parameter holding a mutable sequence of int values, serialized as a comma-separated string."""

    def __init__(self, value: Sequence[int], parent: IntegerSequenceParameter | None) -> None:
        super().__init__(list(value), parent)

    def __iter__(self) -> Iterator[int]:
        return iter(self._value)

    def __getitem__(self, index: int) -> int:
        return self._value[index]

    def __setitem__(self, index: int, value: int) -> None:
        if self._value[index] != value:
            self._value[index] = value
            self.notify_observers()

    def __delitem__(self, index: int) -> None:
        del self._value[index]
        self.notify_observers()

    def insert(self, index: int, value: int) -> None:
        self._value.insert(index, value)
        self.notify_observers()

    def __len__(self) -> int:
        return len(self._value)

    def set_value(self, value: Sequence[int], *, notify: bool = True) -> None:
        if self._value != value:
            self._value = list(value)

            if notify:
                self.notify_observers()

    def get_value_as_string(self) -> str:
        return ','.join(repr(value) for value in self)

    def set_value_from_string(self, value: str) -> None:
        new_value: list[int] = list()

        for xstr in value.split(','):
            if xstr:
                new_value.append(int(xstr))

        self.set_value(new_value)

    def copy(self) -> IntegerSequenceParameter:
        return IntegerSequenceParameter(self.get_value(), self)


class RealSequenceParameter(ParameterBase[MutableSequence[float]]):
    """Parameter holding a mutable sequence of float values, serialized as a comma-separated string."""

    def __init__(self, value: Sequence[float], parent: RealSequenceParameter | None) -> None:
        super().__init__(list(value), parent)

    def __iter__(self) -> Iterator[float]:
        return iter(self._value)

    def __getitem__(self, index: int) -> float:
        return self._value[index]

    def __setitem__(self, index: int, value: float) -> None:
        if self._value[index] != value:
            self._value[index] = value
            self.notify_observers()

    def __delitem__(self, index: int) -> None:
        del self._value[index]
        self.notify_observers()

    def insert(self, index: int, value: float) -> None:
        self._value.insert(index, value)
        self.notify_observers()

    def __len__(self) -> int:
        return len(self._value)

    def set_value(self, value: Sequence[float], *, notify: bool = True) -> None:
        if self._value != value:
            self._value = list(value)

            if notify:
                self.notify_observers()

    def get_value_as_string(self) -> str:
        return ','.join(repr(value) for value in self)

    def set_value_from_string(self, value: str) -> None:
        tmp: list[float] = list()

        for xstr in value.split(','):
            if xstr:
                try:
                    x = float(xstr)
                except ValueError:
                    x = float('nan')

                tmp.append(x)

        self.set_value(tmp)

    def copy(self) -> RealSequenceParameter:
        return RealSequenceParameter(self.get_value(), self)


class ComplexSequenceParameter(ParameterBase[MutableSequence[complex]]):
    """Parameter holding a mutable sequence of complex values, serialized as a comma-separated string."""

    def __init__(self, value: Sequence[complex], parent: ComplexSequenceParameter | None) -> None:
        super().__init__(list(value), parent)

    def __iter__(self) -> Iterator[complex]:
        return iter(self._value)

    def __getitem__(self, index: int) -> complex:
        return self._value[index]

    def __setitem__(self, index: int, value: complex) -> None:
        if self._value[index] != value:
            self._value[index] = value
            self.notify_observers()

    def __delitem__(self, index: int) -> None:
        del self._value[index]
        self.notify_observers()

    def insert(self, index: int, value: complex) -> None:
        self._value.insert(index, value)
        self.notify_observers()

    def __len__(self) -> int:
        return len(self._value)

    def set_value(self, value: Sequence[complex], *, notify: bool = True) -> None:
        if self._value != value:
            self._value = list(value)

            if notify:
                self.notify_observers()

    def get_value_as_string(self) -> str:
        return ','.join(repr(value) for value in self)

    def set_value_from_string(self, value: str) -> None:
        tmp: list[complex] = list()

        for xstr in value.split(','):
            if xstr:
                try:
                    x = complex(xstr)
                except ValueError:
                    x = float('nan') * 1j

                tmp.append(x)

        self.set_value(tmp)

    def copy(self) -> ComplexSequenceParameter:
        return ComplexSequenceParameter(self.get_value(), self)


class ParameterGroup(Observable, Observer):
    """Hierarchical container of named Parameters and nested ParameterGroups."""

    def __init__(self) -> None:
        super().__init__()
        self._parameters: dict[str, Parameter[Any]] = dict()
        self._groups: dict[str, ParameterGroup] = dict()

    def parameters(self) -> Mapping[str, Parameter[Any]]:
        return self._parameters

    def _add_parameter(self, name: str, parameter: Parameter[Any]) -> None:
        if self._parameters.setdefault(name, parameter) is parameter:
            parameter.add_observer(self)
        else:
            raise ValueError(f'Parameter "{name}" already exists!')

    def create_string_parameter(self, name: str, value: str) -> StringParameter:
        parameter = StringParameter(value, parent=None)
        self._add_parameter(name, parameter)
        return parameter

    def create_path_parameter(self, name: str, value: Path) -> PathParameter:
        parameter = PathParameter(value, parent=None)
        self._add_parameter(name, parameter)
        return parameter

    def create_uuid_parameter(self, name: str, value: UUID) -> UUIDParameter:
        parameter = UUIDParameter(value, parent=None)
        self._add_parameter(name, parameter)
        return parameter

    def create_boolean_parameter(self, name: str, value: bool) -> BooleanParameter:
        parameter = BooleanParameter(value, parent=None)
        self._add_parameter(name, parameter)
        return parameter

    def create_integer_parameter(
        self, name: str, value: int, *, minimum: int | None = None, maximum: int | None = None
    ) -> IntegerParameter:
        parameter = IntegerParameter(value, parent=None, minimum=minimum, maximum=maximum)
        self._add_parameter(name, parameter)
        return parameter

    def create_integer_sequence_parameter(
        self, name: str, value: Sequence[int]
    ) -> IntegerSequenceParameter:
        parameter = IntegerSequenceParameter(value, parent=None)
        self._add_parameter(name, parameter)
        return parameter

    def create_real_parameter(
        self, name: str, value: float, *, minimum: float | None = None, maximum: float | None = None
    ) -> RealParameter:
        parameter = RealParameter(value, parent=None, minimum=minimum, maximum=maximum)
        self._add_parameter(name, parameter)
        return parameter

    def create_real_sequence_parameter(
        self, name: str, value: Sequence[float]
    ) -> RealSequenceParameter:
        parameter = RealSequenceParameter(value, parent=None)
        self._add_parameter(name, parameter)
        return parameter

    def create_complex_sequence_parameter(
        self, name: str, value: Sequence[complex]
    ) -> ComplexSequenceParameter:
        parameter = ComplexSequenceParameter(value, parent=None)
        self._add_parameter(name, parameter)
        return parameter

    def groups(self) -> Mapping[str, ParameterGroup]:
        return self._groups

    def _add_group(self, name: str, group: ParameterGroup, *, observe: bool = False) -> None:
        if self._groups.setdefault(name, group) is group:
            if observe:
                group.add_observer(self)
        else:
            raise ValueError(f'Group "{name}" already exists!')

    def _remove_group(self, name: str) -> None:
        try:
            group = self._groups.pop(name)
        except KeyError:
            pass
        else:
            group.remove_observer(self)

    def create_group(self, name: str) -> ParameterGroup:
        group = ParameterGroup()
        self._add_group(name, group)
        return group

    def get_group(self, name: str) -> ParameterGroup:
        return self._groups[name]

    def _update(self, observable: Observable) -> None:
        if observable in self._parameters.values():
            self.notify_observers()
        elif observable in self._groups.values():
            self.notify_observers()



from __future__ import annotations
from decimal import Decimal, ROUND_FLOOR

from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QComboBox, QHBoxLayout, QWidget

from .decimal_line_edit import DecimalLineEdit


class LengthWidget(QWidget):
    length_changed = pyqtSignal(Decimal)

    def __init__(self, is_signed: bool, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.length_m = Decimal()
        self.line_edit = DecimalLineEdit.create_instance(is_signed=is_signed)
        self.units_combo_box = QComboBox()

        if not is_signed:
            self.line_edit.set_minimum(Decimal())

        self.line_edit.value_changed.connect(self._set_length_m_from_widgets)

        self.units_combo_box.addItem('m', 0)
        self.units_combo_box.addItem('mm', -3)
        self.units_combo_box.addItem('\u00b5m', -6)
        self.units_combo_box.addItem('nm', -9)
        self.units_combo_box.addItem('\u212b', -10)
        self.units_combo_box.addItem('pm', -12)
        self.units_combo_box.activated.connect(self._update_display)

        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.line_edit)
        layout.addWidget(self.units_combo_box)
        self.setLayout(layout)

    def is_read_only(self) -> bool:
        return self.line_edit.is_read_only()

    def set_read_only(self, enable: bool) -> None:
        self.line_edit.set_read_only(enable)

    def get_length_m(self) -> Decimal:
        return self.length_m

    def set_length_m(self, length_m: Decimal) -> None:
        self.length_m = length_m

        if not length_m.is_zero():
            exponent = 3 * int((abs(length_m).log10() / 3).to_integral_exact(rounding=ROUND_FLOOR))
            index = self.units_combo_box.findData(exponent)

            if index != -1:
                self.units_combo_box.setCurrentIndex(index)

        self._update_display()

    @property
    def _scale_to_meters(self) -> Decimal:
        exponent = self.units_combo_box.currentData()
        return Decimal(f'1e{exponent:+d}')

    def _set_length_m_from_widgets(self, magnitude: Decimal) -> None:
        self.length_m = magnitude * self._scale_to_meters
        self.length_changed.emit(self.length_m)

    def _update_display(self) -> None:
        length_in_display_units = self.length_m / self._scale_to_meters
        self.line_edit.set_value(length_in_display_units)



from __future__ import annotations
from enum import IntEnum
from pathlib import Path
from typing import Final
import csv

import numpy

from ptychodus.api.probe_positions import (
    ProbePositionSequence,
    ProbePositionFileReader,
    ProbePosition,
    ProbePositionParseError,
)
from .aps33id_velociprobe_diffraction_file import VelociprobeDiffractionFileReader

__all__ = [
    'VelociprobePositionFileReader',
]


class VelociprobePositionFileColumn(IntEnum):
    X = 1
    LASER_INTERFEROMETER_Y = 2
    POSITION_ENCODER_Y = 5
    TRIGGER = 7


class VelociprobePositionFileReader(ProbePositionFileReader):
    ONE_NANOMETER_M: Final[float] = 1.0e-9

    def __init__(self, diffraction_reader: VelociprobeDiffractionFileReader, y_column: int) -> None:
        self._diffraction_reader = diffraction_reader
        self._y_column = y_column

    @classmethod
    def create_laser_interferometer_instance(
        cls, diffraction_reader: VelociprobeDiffractionFileReader
    ) -> VelociprobePositionFileReader:
        return cls(diffraction_reader, VelociprobePositionFileColumn.LASER_INTERFEROMETER_Y)

    @classmethod
    def create_position_encoder_instance(
        cls, diffraction_reader: VelociprobeDiffractionFileReader
    ) -> VelociprobePositionFileReader:
        return cls(diffraction_reader, VelociprobePositionFileColumn.POSITION_ENCODER_Y)

    def _apply_transform(self, positions: ProbePositionSequence) -> ProbePositionSequence:
        stage_rotation_rad = numpy.deg2rad(self._diffraction_reader.stage_rotation_deg)
        stage_rotation_cos = numpy.cos(stage_rotation_rad)

        x_mean = sum(p.coordinate_x_m for p in positions) / len(positions)
        y_mean = sum(p.coordinate_y_m for p in positions) / len(positions)
        point_list: list[ProbePosition] = list()

        for untransformed_point in positions:
            point = ProbePosition(
                untransformed_point.index,
                (untransformed_point.coordinate_x_m - x_mean) * stage_rotation_cos,
                (untransformed_point.coordinate_y_m - y_mean),
            )
            point_list.append(point)

        return ProbePositionSequence(point_list)

    def read(self, file_path: Path) -> ProbePositionSequence:
        point_list: list[ProbePosition] = list()
        minimum_column_count = max(col.value for col in VelociprobePositionFileColumn) + 1

        with file_path.open(newline='') as csv_file:
            csv_reader = csv.reader(csv_file, delimiter=',')

            for row in csv_reader:
                if row[0].startswith('#'):
                    continue

                if len(row) < minimum_column_count:
                    raise ProbePositionParseError('Bad number of columns!')

                trigger = int(row[VelociprobePositionFileColumn.TRIGGER])
                x_nm = int(row[VelociprobePositionFileColumn.X])
                y_nm = int(row[self._y_column])

                if self._y_column == VelociprobePositionFileColumn.POSITION_ENCODER_Y:
                    y_nm = -y_nm

                point = ProbePosition(
                    trigger,
                    x_nm * self.ONE_NANOMETER_M,
                    y_nm * self.ONE_NANOMETER_M,
                )
                point_list.append(point)

        raw_positions = ProbePositionSequence(point_list)

        return self._apply_transform(raw_positions)



from __future__ import annotations
from collections.abc import Sequence

import numpy
from scipy.ndimage import gaussian_filter

from ptychodus.api.common import TWO_PI_J
from ptychodus.api.object import Object, ObjectGeometryProvider

from .builder import ObjectBuilder
from .settings import ObjectSettings


class RandomObjectBuilder(ObjectBuilder):
    def __init__(self, rng: numpy.random.Generator, settings: ObjectSettings) -> None:
        super().__init__(settings, 'random')
        self._rng = rng
        self._settings = settings

        self.amplitude_mean = settings.amplitude_mean.copy()
        self._add_parameter('amplitude_mean', self.amplitude_mean)
        self.amplitude_deviation = settings.amplitude_deviation.copy()
        self._add_parameter('amplitude_deviation', self.amplitude_deviation)
        self.phase_deviation_turns = settings.phase_deviation_turns.copy()
        self._add_parameter('phase_deviation_turns', self.phase_deviation_turns)
        self.blur_deviation_px = settings.blur_deviation_px.copy()
        self._add_parameter('blur_deviation_px', self.blur_deviation_px)

    def copy(self) -> RandomObjectBuilder:
        builder = RandomObjectBuilder(self._rng, self._settings)

        for key, value in self.parameters().items():
            builder.parameters()[key].set_value(value.get_value())

        return builder

    def build(
        self,
        geometry_provider: ObjectGeometryProvider,
        layer_spacing_m: Sequence[float],
    ) -> Object:
        geometry = geometry_provider.get_object_geometry()
        object_shape = (1, geometry.height_px, geometry.width_px)

        amplitude = self._rng.normal(
            self.amplitude_mean.get_value(),
            self.amplitude_deviation.get_value(),
            object_shape,
        )
        phase = self._rng.normal(
            0.0,
            self.phase_deviation_turns.get_value(),
            object_shape,
        )

        blur_sigma = self.blur_deviation_px.get_value()

        if blur_sigma > 0.0:
            # TODO account for pixel aspect ratio in blur calculation
            amplitude = gaussian_filter(amplitude, sigma=blur_sigma)
            phase = gaussian_filter(phase, sigma=blur_sigma)

        array = amplitude * numpy.exp(TWO_PI_J * phase)

        return self._create_object(
            array=array.astype('complex'),
            layer_spacing_m=layer_spacing_m,
            pixel_geometry=geometry.get_pixel_geometry(),
            center=geometry.get_center(),
        )



from collections.abc import Sequence
from typing import overload
import logging

from ptychodus.api.observer import ObservableSequence
from ptychodus.api.product import LossValue

from .item import ProductRepositoryItem, ProductRepositoryObserver
from .metadata import MetadataRepositoryItem
from .object import ObjectRepositoryItem
from .probe import ProbeRepositoryItem
from .repository import ProductRepository
from .probe_positions import ProbePositionsRepositoryItem

logger = logging.getLogger(__name__)


class ObjectRepository(ObservableSequence[ObjectRepositoryItem], ProductRepositoryObserver):
    def __init__(self, repository: ProductRepository) -> None:
        super().__init__()
        self._repository = repository
        self._repository.add_observer(self)

    def get_name(self, index: int) -> str:
        return self._repository[index].get_name()

    def set_name(self, index: int, name: str) -> None:
        self._repository[index].set_name(name)

    @overload
    def __getitem__(self, index: int) -> ObjectRepositoryItem: ...

    @overload
    def __getitem__(self, index: slice) -> Sequence[ObjectRepositoryItem]: ...

    def __getitem__(
        self, index: int | slice
    ) -> ObjectRepositoryItem | Sequence[ObjectRepositoryItem]:
        if isinstance(index, slice):
            return [item.get_object_item() for item in self._repository[index]]
        else:
            return self._repository[index].get_object_item()

    def __len__(self) -> int:
        return len(self._repository)

    def handle_item_inserted(self, index: int, item: ProductRepositoryItem) -> None:
        self.notify_observers_item_inserted(index, item.get_object_item())

    def handle_metadata_changed(self, index: int, item: MetadataRepositoryItem) -> None:
        pass

    def handle_probe_positions_changed(
        self, index: int, item: ProbePositionsRepositoryItem
    ) -> None:
        pass

    def handle_probe_changed(self, index: int, item: ProbeRepositoryItem) -> None:
        pass

    def handle_object_changed(self, index: int, item: ObjectRepositoryItem) -> None:
        self.notify_observers_item_changed(index, item)

    def handle_losses_changed(self, index: int, losses: Sequence[LossValue]) -> None:
        pass

    def handle_item_removed(self, index: int, item: ProductRepositoryItem) -> None:
        self.notify_observers_item_removed(index, item.get_object_item())



from pathlib import Path
from typing import Final
import logging

import h5py
import numpy

from ptychodus.api.geometry import ImageExtent, PixelGeometry
from ptychodus.api.diffraction import (
    DiffractionDataset,
    DiffractionFileReader,
    DiffractionMetadata,
    SimpleDiffractionDataset,
)
from ptychodus.api.plugins import PluginRegistry

from .h5_diffraction_file import H5DiffractionPatternArray, H5DiffractionFileTreeBuilder

logger = logging.getLogger(__name__)


class CSAXSDiffractionFileReader(DiffractionFileReader):
    ONE_MICRON_M: Final[float] = 1e-6
    ONE_MILLIMETER_M: Final[float] = 1e-3

    def __init__(self) -> None:
        self._data_path = '/entry/data/data'
        self._tree_builder = H5DiffractionFileTreeBuilder()

    def read(self, file_path: Path) -> DiffractionDataset:
        with h5py.File(file_path, 'r') as h5_file:
            data = h5_file[self._data_path]

            if isinstance(data, h5py.Dataset):
                num_patterns, detector_height, detector_width = data.shape

                distance_mm = h5_file['/entry/instrument/monochromator/distance']
                x_pixel_size_um = h5_file['/entry/instrument/eiger_4/x_pixel_size']
                y_pixel_size_um = h5_file['/entry/instrument/eiger_4/y_pixel_size']
                energy_keV = h5_file['/entry/instrument/monochromator/energy']  # noqa: N806

                metadata = DiffractionMetadata(
                    num_patterns_per_array=[num_patterns],
                    pattern_dtype=data.dtype,
                    detector_distance_m=abs(float(distance_mm[()])) * self.ONE_MILLIMETER_M,
                    detector_extent=ImageExtent(detector_width, detector_height),
                    detector_pixel_geometry=PixelGeometry(
                        width_m=float(x_pixel_size_um[()]) * self.ONE_MICRON_M,
                        height_m=float(y_pixel_size_um[()]) * self.ONE_MICRON_M,
                    ),
                    probe_energy_eV=1000 * float(energy_keV[()]),
                    file_path=file_path,
                )
                contents_tree = self._tree_builder.build(h5_file)
                array = H5DiffractionPatternArray(
                    label=file_path.stem,
                    indexes=numpy.arange(num_patterns),
                    file_path=file_path,
                    data_path=self._data_path,
                )

                return SimpleDiffractionDataset(metadata, contents_tree, [array])
            else:
                raise ValueError(f'Expected {self._data_path} to be a dataset; got {type(data)}.')


def register_plugins(registry: PluginRegistry) -> None:
    registry.diffraction_file_readers.register_plugin(
        CSAXSDiffractionFileReader(),
        simple_name='SLS_cSAXS',
        display_name='SLS cSAXS Files (*.h5 *.hdf5)',
    )



from abc import ABC, abstractmethod
from collections.abc import Iterator

import numpy

from ptychodus.api.common import RealArrayType
from ptychodus.api.observer import Observable, Observer
from ptychodus.api.parametric import Parameter
from ptychodus.api.plugins import PluginChooser

__all__ = [
    'ScalarTransformation',
    'ScalarTransformationParameter',
]


class ScalarTransformation(ABC):
    @abstractmethod
    def decorate_text(self, text: str) -> str:
        pass

    @abstractmethod
    def __call__(self, array: RealArrayType) -> RealArrayType:
        pass


class IdentityScalarTransformation(ScalarTransformation):
    def decorate_text(self, text: str) -> str:
        return text

    def __call__(self, array: RealArrayType) -> RealArrayType:
        return array


class SquareRootScalarTransformation(ScalarTransformation):
    def decorate_text(self, text: str) -> str:
        return f'$\\sqrt{{\\mathrm{{{text}}}}}$'

    def __call__(self, array: RealArrayType) -> RealArrayType:
        nil = numpy.zeros_like(array)
        return numpy.sqrt(array, out=nil, where=(array > 0))


class Log2ScalarTransformation(ScalarTransformation):
    def decorate_text(self, text: str) -> str:
        return f'$\\log_2{{\\left(\\mathrm{{{text}}}\\right)}}$'

    def __call__(self, array: RealArrayType) -> RealArrayType:
        nil = numpy.zeros_like(array)
        return numpy.log2(array, out=nil, where=(array > 0))


class LogScalarTransformation(ScalarTransformation):
    def decorate_text(self, text: str) -> str:
        return f'$\\ln{{\\left(\\mathrm{{{text}}}\\right)}}$'

    def __call__(self, array: RealArrayType) -> RealArrayType:
        nil = numpy.zeros_like(array)
        return numpy.log(array, out=nil, where=(array > 0))


class Log10ScalarTransformation(ScalarTransformation):
    def decorate_text(self, text: str) -> str:
        return f'$\\log_{{10}}{{\\left(\\mathrm{{{text}}}\\right)}}$'

    def __call__(self, array: RealArrayType) -> RealArrayType:
        nil = numpy.zeros_like(array)
        return numpy.log10(array, out=nil, where=(array > 0))


class ScalarTransformationParameter(Parameter[str], Observer):
    def __init__(self) -> None:
        super().__init__()
        self._chooser = PluginChooser[ScalarTransformation]()
        self._chooser.register_plugin(
            IdentityScalarTransformation(),
            display_name='Identity',
        )
        self._chooser.register_plugin(
            SquareRootScalarTransformation(),
            simple_name='sqrt',
            display_name='Square Root',
        )
        self._chooser.register_plugin(
            Log2ScalarTransformation(),
            simple_name='log2',
            display_name='Logarithm (Base 2)',
        )

        self._chooser.register_plugin(
            LogScalarTransformation(),
            simple_name='ln',
            display_name='Natural Logarithm',
        )
        self._chooser.register_plugin(
            Log10ScalarTransformation(),
            simple_name='log10',
            display_name='Logarithm (Base 10)',
        )
        self.set_value('Identity')
        self._chooser.add_observer(self)

    def choices(self) -> Iterator[str]:
        for plugin in self._chooser:
            yield plugin.display_name

    def get_value(self) -> str:
        return self._chooser.get_current_plugin().display_name

    def set_value(self, value: str, *, notify: bool = True) -> None:
        self._chooser.set_current_plugin(value)

    def get_value_as_string(self) -> str:
        return self.get_value()

    def set_value_from_string(self, value: str) -> None:
        self.set_value(value)

    def copy(self) -> Parameter[str]:
        parameter = ScalarTransformationParameter()
        parameter.set_value(self.get_value())
        return parameter

    def decorate_text(self, text: str) -> str:
        return self._chooser.get_current_plugin().strategy.decorate_text(text)

    def transform(self, values: RealArrayType) -> RealArrayType:
        return self._chooser.get_current_plugin().strategy(values)

    def _update(self, observable: Observable) -> None:
        if observable is self._chooser:
            self.notify_observers()



import logging

from ptychodus.api.observer import Observable, Observer
from ptychodus.api.diffraction import (
    BadPixelsFileReader,
    DiffractionFileReader,
    DiffractionFileWriter,
)
from ptychodus.api.plugins import PluginChooser
from ptychodus.api.settings import SettingsRegistry

from ..task_manager import TaskManager
from .api import DiffractionAPI
from .bad_pixels import BadPixelsProvider
from .dataset import AssembledDiffractionDataset
from .settings import DetectorSettings, DiffractionSettings
from .sizer import PatternSizer

logger = logging.getLogger(__name__)


class DiffractionCore(Observer):
    def __init__(
        self,
        task_manager: TaskManager,
        settings_registry: SettingsRegistry,
        bad_pixels_file_reader_chooser: PluginChooser[BadPixelsFileReader],
        file_reader_chooser: PluginChooser[DiffractionFileReader],
        file_writer_chooser: PluginChooser[DiffractionFileWriter],
        reinit_observable: Observable,
    ) -> None:
        super().__init__()
        self.detector_settings = DetectorSettings(settings_registry)
        self.diffraction_settings = DiffractionSettings(settings_registry)
        self.pattern_sizer = PatternSizer(self.detector_settings, self.diffraction_settings)
        self.bad_pixels_provider = BadPixelsProvider(self.detector_settings)
        self.dataset = AssembledDiffractionDataset(
            self.diffraction_settings, self.pattern_sizer, self.bad_pixels_provider, task_manager
        )
        self.diffraction_api = DiffractionAPI(
            self.diffraction_settings,
            self.detector_settings,
            self.bad_pixels_provider,
            self.dataset,
            bad_pixels_file_reader_chooser,
            file_reader_chooser,
            file_writer_chooser,
        )

        bad_pixels_file_reader_chooser.synchronize_with_parameter(
            self.detector_settings.bad_pixels_file_type
        )
        file_reader_chooser.synchronize_with_parameter(self.diffraction_settings.file_type)
        file_writer_chooser.set_current_plugin(self.diffraction_settings.file_type.get_value())

        self._reinit_observable = reinit_observable
        reinit_observable.add_observer(self)

    def _update(self, observable: Observable) -> None:
        if observable is self._reinit_observable:
            self.diffraction_api.open_bad_pixels(
                file_path=self.detector_settings.bad_pixels_file_path.get_value(),
                file_type=self.detector_settings.bad_pixels_file_type.get_value(),
            )
            self.diffraction_api.open_patterns(
                file_path=self.diffraction_settings.file_path.get_value(),
                file_type=self.diffraction_settings.file_type.get_value(),
            )
            self.diffraction_api.load_all_arrays()



"""Probe position (scan point) data structures and file I/O plugin interfaces."""

from __future__ import annotations
from abc import ABC, abstractmethod
from collections.abc import Sequence
from dataclasses import dataclass
from pathlib import Path
from typing import overload

import numpy


@dataclass(frozen=True)
class ProbePosition:
    """Probe position with its scan index and (x, y) physical coordinates in meters."""

    index: int
    coordinate_x_m: float
    coordinate_y_m: float


@dataclass(frozen=True)
class ScanBoundingBox:
    """Axis-aligned bounding box enclosing a set of probe positions."""

    minimum_x_m: float
    maximum_x_m: float
    minimum_y_m: float
    maximum_y_m: float

    @property
    def width_m(self) -> float:
        return self.maximum_x_m - self.minimum_x_m

    @property
    def height_m(self) -> float:
        return self.maximum_y_m - self.minimum_y_m

    @property
    def center_x_m(self) -> float:
        return self.minimum_x_m + self.width_m / 2.0

    @property
    def center_y_m(self) -> float:
        return self.minimum_y_m + self.height_m / 2.0

    def hull(self, bbox: ScanBoundingBox) -> ScanBoundingBox:
        return ScanBoundingBox(
            minimum_x_m=min(self.minimum_x_m, bbox.minimum_x_m),
            maximum_x_m=max(self.maximum_x_m, bbox.maximum_x_m),
            minimum_y_m=min(self.minimum_y_m, bbox.minimum_y_m),
            maximum_y_m=max(self.maximum_y_m, bbox.maximum_y_m),
        )


class ProbePositionSequence(Sequence[ProbePosition]):
    """Memory-efficient sequence of ProbePosition objects backed by numpy arrays."""

    def __init__(self, point_seq: Sequence[ProbePosition] | None = None) -> None:
        indexes: list[int] = []
        coordinates_m: list[float] = []

        if point_seq is not None:
            for point in point_seq:
                indexes.append(point.index)
                coordinates_m.append(point.coordinate_y_m)
                coordinates_m.append(point.coordinate_x_m)

        self._indexes = numpy.array(indexes)
        self._coordinates_m = numpy.reshape(coordinates_m, (-1, 2))

    def copy(self) -> ProbePositionSequence:
        seq = ProbePositionSequence()
        seq._indexes = self._indexes.copy()
        seq._coordinates_m = self._coordinates_m.copy()
        return seq

    @overload
    def __getitem__(self, index: int) -> ProbePosition: ...

    @overload
    def __getitem__(self, index: slice) -> Sequence[ProbePosition]: ...

    def __getitem__(self, index: int | slice) -> ProbePosition | Sequence[ProbePosition]:
        if isinstance(index, slice):
            return [self[idx] for idx in range(index.start, index.stop, index.step)]

        return ProbePosition(
            index=self._indexes[index],
            coordinate_x_m=self._coordinates_m[index, -1],
            coordinate_y_m=self._coordinates_m[index, -2],
        )

    def __len__(self) -> int:
        return self._indexes.size

    @property
    def nbytes(self) -> int:
        return self._indexes.nbytes + self._coordinates_m.nbytes

    def __repr__(self) -> str:
        return f'{self._coordinates_m.dtype}{self._coordinates_m.shape}'


class ProbePositionParseError(Exception):
    """Raised when a probe position file cannot be parsed."""

    pass


class ProbePositionFileReader(ABC):
    """Plugin interface for reading probe position files."""

    @abstractmethod
    def read(self, file_path: Path) -> ProbePositionSequence:
        """Read probe positions from file."""
        pass


class ProbePositionFileWriter(ABC):
    """Plugin interface for writing probe position files."""

    @abstractmethod
    def write(self, file_path: Path, positions: ProbePositionSequence) -> None:
        """Write probe positions to file."""
        pass



from PyQt5.QtGui import QValidator
from PyQt5.QtWidgets import QSpinBox, QWidget

from ptychodus.api.observer import Observable, Observer
from ptychodus.api.parametric import IntegerParameter

from ..model.ptychopinn.core import PtychoPINNReconstructorLibrary
from .data import FileDialogFactory
from .parametric import ParameterViewBuilder, ParameterViewController
from .processing import ReconstructorViewControllerFactory


class PowerTwoSpinBox(QSpinBox):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)

    def stepBy(self, steps: int) -> None:  # noqa: N802
        if steps < 0:
            self.setValue(self.value() // (1 << -steps))
        elif steps > 0:
            self.setValue(self.value() * (1 << steps))

    def validate(self, input: str, pos: int) -> tuple[QValidator.State, str, int]:
        try:
            value = int(input)
        except ValueError:
            pass
        else:
            if value > 0:
                is_pow2 = (value & (value - 1)) == 0

                if is_pow2:
                    return QValidator.State.Acceptable, input, pos

        return QValidator.State.Intermediate, input, pos


class PowerTwoSpinBoxParameterViewController(ParameterViewController, Observer):
    def __init__(self, parameter: IntegerParameter, *, tool_tip: str = '') -> None:
        super().__init__()
        self._parameter = parameter
        self._widget = PowerTwoSpinBox()

        if tool_tip:
            self._widget.setToolTip(tool_tip)

        self._sync_model_to_view()
        self._widget.valueChanged.connect(parameter.set_value)
        parameter.add_observer(self)

    def get_widget(self) -> QWidget:
        return self._widget

    def _sync_model_to_view(self) -> None:
        minimum = self._parameter.get_minimum()
        maximum = self._parameter.get_maximum()

        if minimum is None:
            raise ValueError('Minimum not provided!')

        if maximum is None:
            raise ValueError('Maximum not provided!')

        self._widget.blockSignals(True)
        self._widget.setRange(minimum, maximum)
        self._widget.setValue(self._parameter.get_value())
        self._widget.blockSignals(False)

    def _update(self, observable: Observable) -> None:
        if observable is self._parameter:
            self._sync_model_to_view()


class PtychoPINNViewControllerFactory(ReconstructorViewControllerFactory):
    def __init__(
        self, model: PtychoPINNReconstructorLibrary, file_dialog_factory: FileDialogFactory
    ) -> None:
        super().__init__()
        self._model = model
        self._file_dialog_factory = file_dialog_factory

    @property
    def name(self) -> str:
        return 'PtychoPINN'

    def create_view_controller(self, reconstructor_name: str) -> QWidget:
        is_pinn = reconstructor_name.lower() == 'pinn'
        builder = ParameterViewBuilder(self._file_dialog_factory)
        enumerators = self._model.enumerators

        model_group = 'Model'
        model_settings = self._model.model_settings
        builder.add_spin_box(
            model_settings.n_filters_scale,
            'Num. Filters Scale Factor:',
            tool_tip='Scale factor for number of filters',
            group=model_group,
        )

        if is_pinn:
            builder.add_spin_box(
                model_settings.gridsize,
                'Grid Size:',
                tool_tip='Controls number of images processed per solution region (e.g., gridsize=2 means 2^2=4 images at a time)',
                group=model_group,
            )
            builder.add_combo_box(
                model_settings.amp_activation,
                enumerators.get_amp_activations(),
                'Amplitude Activation Function:',
                group=model_group,
            )
            builder.add_check_box(
                model_settings.object_big,
                'Object Big',
                group=model_group,
                tool_tip='Enables a separate real-space reconstruction for each input diffraction image '
                'and an averaging / overlap constraint step. If False, no explicit averaging is performed '
                'and the decoders return a single real space image instead of `gridsize**2` images. '
                'Typically left True.',
            )
            builder.add_check_box(
                model_settings.probe_big,
                'Probe Big',
                group=model_group,
                tool_tip='If True, enables a low-resolution reconstruction of the outer region of the NxN real-space grid. '
                'This technically violates the zero-padding / oversampling condition, '
                'but may be needed if the probe illumination has wide tails. '
                'Has no effect unless pad_object is True.',
            )
            builder.add_check_box(
                model_settings.probe_mask,
                'Probe Mask',
                group=model_group,
                tool_tip='Whether to apply circular mask to the probe function. '
                "If toggling this changes the reconstruction, it's likely that there are edge / real space truncation artifacts. "
                'Should be used with pad_object = False.',
            )
            builder.add_check_box(
                model_settings.pad_object,
                'Pad Object',
                group=model_group,
                tool_tip='Whether to reconstruct the full real space grid (False) or restrict to N/2 x N/2 (True). '
                'True strictly enforces the necessary reciprocal space oversampling, '
                'but may cause truncation issues for probe amplitudes with long tails. '
                'This truncation can be mitigated by setting probe_big, '
                'which uses a small number of CNN filters to generate a low-resolution reconstruction of the outer region. '
                'Typically left True.',
            )
            builder.add_decimal_line_edit(
                model_settings.probe_scale,
                'Probe Scale Factor:',
                group=model_group,
                tool_tip='Scaling factor for the probe amplitude. ',
            )
            builder.add_decimal_line_edit(
                model_settings.gaussian_smoothing_sigma,
                'Gaussian Smoothing Sigma:',
                group=model_group,
                tool_tip='Standard deviation for Gaussian smoothing of probe illumination.  '
                'Increase from 0 to reduce noise / artifacts at cost of resolution.  '
                'Beware that abusing this can cause convergence issues.',
            )

        inference_group = 'Inference'
        inference_settings = self._model.inference_settings
        builder.add_spin_box(
            inference_settings.n_nearest_neighbors, 'Number of Neighbors:', group=inference_group
        )
        builder.add_spin_box(
            inference_settings.n_samples, 'Number of Samples:', group=inference_group
        )

        training_group = 'Training'
        training_settings = self._model.training_settings
        builder.add_view_controller(
            PowerTwoSpinBoxParameterViewController(training_settings.batch_size),
            'Batch Size:',
            group=training_group,
        )
        builder.add_spin_box(training_settings.nepochs, 'Number of Epochs:', group=training_group)

        if is_pinn:
            builder.add_decimal_slider(
                training_settings.mae_weight, 'Weight for MAE loss:', group=training_group
            )
            builder.add_decimal_slider(
                training_settings.nll_weight, 'Weight for NLL loss:', group=training_group
            )
            builder.add_decimal_slider(
                training_settings.realspace_mae_weight,
                'Realspace MAE Weight:',
                group=training_group,
            )
            builder.add_decimal_slider(
                training_settings.realspace_weight, 'Realspace Weight:', group=training_group
            )
            builder.add_check_box(
                training_settings.positions_provided,
                'Positions Provided',
                group=training_group,
            )
            builder.add_check_box(
                training_settings.probe_trainable,
                'Probe Trainable',
                group=training_group,
                tool_tip='Optimizes the probe function during training. Experimental feature.',
            )
            builder.add_check_box(
                training_settings.intensity_scale_trainable,
                'Intensity Scale Trainable',
                group=training_group,
                tool_tip="Optimize the model's internal amplitude scaling factor during training. "
                'Typically left True.',
            )

        return builder.build_widget()



import getpass

from ptychodus.api.observer import Observable, Observer
from ptychodus.api.settings import SettingsRegistry


class ArgoSettings(Observable, Observer):
    def __init__(self, registry: SettingsRegistry) -> None:
        super().__init__()
        self._group = registry.create_group('Argo')
        self._group.add_observer(self)

        self.user = self._group.create_string_parameter('User', getpass.getuser())
        self.chat_endpoint_url = self._group.create_string_parameter(
            'ChatEndpointURL', 'https://apps.inside.anl.gov/argoapi/api/v1/resource/chat/'
        )
        self.chat_model = self._group.create_string_parameter('ChatModel', 'gpt35')
        self.temperature = self._group.create_real_parameter(
            'Temperature', 0.1, minimum=0.0, maximum=2.0
        )
        self.top_p = self._group.create_real_parameter('TopP', 0.9, minimum=0.0, maximum=1.0)
        self.max_tokens = self._group.create_integer_parameter(
            'MaxTokens', 1000, minimum=0, maximum=128000
        )
        self.max_completion_tokens = self._group.create_integer_parameter(
            'MaxCompletionTokens', 1000, minimum=0, maximum=128000
        )
        self.embeddings_endpoint_url = self._group.create_string_parameter(
            'EmbeddingsEndpointURL', 'https://apps.inside.anl.gov/argoapi/api/v1/resource/embed/'
        )
        self.embeddings_model = self._group.create_string_parameter('EmbeddingsModel', 'ada002')

    def _update(self, observable: Observable) -> None:
        if observable is self._group:
            self.notify_observers()



from .builder import FromFileProbePositionsBuilder, FromMemoryProbePositionsBuilder
from .builder_factory import ProbePositionsBuilderFactory
from .cartesian import CartesianProbePositionsBuilder
from .concentric import ConcentricProbePositionsBuilder
from .item import ProbePositionsRepositoryItem
from .item_factory import ProbePositionsRepositoryItemFactory
from .lissajous import LissajousProbePositionsBuilder
from .settings import ProbePositionsSettings
from .spiral import SpiralProbePositionsBuilder
from .transform import ProbePositionTransform

__all__ = [
    'CartesianProbePositionsBuilder',
    'ConcentricProbePositionsBuilder',
    'FromFileProbePositionsBuilder',
    'FromMemoryProbePositionsBuilder',
    'LissajousProbePositionsBuilder',
    'ProbePositionsSettings',
    'ProbePositionsBuilderFactory',
    'ProbePositionsRepositoryItem',
    'ProbePositionsRepositoryItemFactory',
    'ProbePositionTransform',
    'SpiralProbePositionsBuilder',
]



import logging

from PyQt5.QtWidgets import QWizard

from ....model.metadata import MetadataPresenter
from ....model.diffraction import DiffractionSettings, PatternSizer, DiffractionAPI

from ...data import FileDialogFactory
from .files import OpenDatasetWizardFilesViewController
from .metadata import OpenDatasetWizardMetadataViewController
from .patterns import OpenDatasetWizardPatternsViewController

logger = logging.getLogger(__name__)


class OpenDatasetWizardController:
    def __init__(
        self,
        settings: DiffractionSettings,
        sizer: PatternSizer,
        api: DiffractionAPI,
        metadata_presenter: MetadataPresenter,
        file_dialog_factory: FileDialogFactory,
    ) -> None:
        self._api = api
        self._file_view_controller = OpenDatasetWizardFilesViewController(
            settings, api, file_dialog_factory
        )
        self._metadata_view_controller = OpenDatasetWizardMetadataViewController(metadata_presenter)
        self._patterns_view_controller = OpenDatasetWizardPatternsViewController(
            settings, sizer, file_dialog_factory
        )

        self._wizard = QWizard()
        self._wizard.setWindowTitle('Open Dataset')
        self._wizard.addPage(self._file_view_controller.get_widget())
        self._wizard.addPage(self._metadata_view_controller.get_widget())
        self._wizard.addPage(self._patterns_view_controller.get_widget())

        next_button = self._wizard.button(QWizard.WizardButton.NextButton)

        if next_button is None:
            raise ValueError('next_button is None!')
        else:
            next_button.clicked.connect(self._execute_next_button_action)

        finish_button = self._wizard.button(QWizard.WizardButton.FinishButton)

        if finish_button is None:
            raise ValueError('finish_button is None!')
        else:
            finish_button.clicked.connect(self._execute_finish_button_action)

    def _execute_next_button_action(self) -> None:
        page = self._wizard.currentPage()

        if page is self._metadata_view_controller.get_widget():
            self._file_view_controller.open_dataset()
        elif page is self._patterns_view_controller.get_widget():
            self._metadata_view_controller.import_metadata()

    def _execute_finish_button_action(self) -> None:
        self._api.load_all_arrays()

    def open_dataset(self) -> None:
        self._wizard.restart()
        self._file_view_controller.restart()
        self._wizard.show()



import logging

from PyQt5.QtCore import Qt, QModelIndex
from PyQt5.QtGui import QPainter
from PyQt5.QtWidgets import (
    QApplication,
    QStyle,
    QStyleOptionProgressBar,
    QStyledItemDelegate,
    QStyleOptionViewItem,
)

logger = logging.getLogger(__name__)


class ProgressBarItemDelegate(QStyledItemDelegate):
    def paint(
        self, painter: QPainter | None, option: QStyleOptionViewItem, index: QModelIndex
    ) -> None:
        if painter is None:
            return

        text = index.data(Qt.ItemDataRole.DisplayRole)
        progress = index.data(Qt.ItemDataRole.UserRole)

        if progress is None:
            logger.error(f'Bad data at row={index.row()}, col={index.column()}!')
        elif progress >= 0:
            opt = QStyleOptionProgressBar()
            opt.rect = option.rect
            opt.minimum = 0
            opt.maximum = 100
            opt.progress = int(progress)
            opt.text = text
            opt.textVisible = True
            style = QApplication.style()

            if style is not None:
                style.drawControl(QStyle.ControlElement.CE_ProgressBar, opt, painter)



from __future__ import annotations

import numpy

from ptychodus.api.probe import ProbeSequence, ProbeGeometryProvider
from ptychodus.api.propagator import FresnelTransformPropagator, PropagatorParameters

from ...diffraction import DiffractionAPI
from .builder import ProbeSequenceBuilder
from .settings import ProbeSettings


class AveragePatternProbeBuilder(ProbeSequenceBuilder):
    def __init__(
        self,
        settings: ProbeSettings,
        diffraction_api: DiffractionAPI,
    ) -> None:
        super().__init__(settings, 'average_pattern')
        self._settings = settings
        self._diffraction_api = diffraction_api

    def copy(self) -> AveragePatternProbeBuilder:
        return AveragePatternProbeBuilder(self._settings, self._diffraction_api)

    def build(self, geometry_provider: ProbeGeometryProvider) -> ProbeSequence:
        geometry = geometry_provider.get_probe_geometry()
        assembled_data = self._diffraction_api.get_assembled_data()
        detector_intensity = numpy.mean(assembled_data.get_patterns(), axis=0)

        pixel_geometry = geometry_provider.get_detector_pixel_geometry()
        propagator_parameters = PropagatorParameters(
            wavelength_m=geometry_provider.probe_wavelength_m,
            width_px=detector_intensity.shape[-1],
            height_px=detector_intensity.shape[-2],
            pixel_width_m=pixel_geometry.width_m,
            pixel_height_m=pixel_geometry.height_m,
            propagation_distance_m=-geometry_provider.detector_distance_m,
        )
        propagator = FresnelTransformPropagator(propagator_parameters)
        array = propagator.propagate(numpy.sqrt(detector_intensity).astype(complex))

        return ProbeSequence(
            array=self.normalize(array),
            opr_weights=None,
            pixel_geometry=geometry.get_pixel_geometry(),
        )



from __future__ import annotations
from abc import ABC, abstractmethod
from bisect import bisect
from collections.abc import Sequence
from pathlib import Path
from typing import overload
import logging
import tempfile

import numpy

from ptychodus.api.common import BYTES_PER_MEGABYTE
from ptychodus.api.geometry import ImageExtent
from ptychodus.api.diffraction import (
    BadPixels,
    DiffractionArray,
    DiffractionDataset,
    DiffractionIndexes,
    DiffractionMetadata,
    DiffractionPattern,
    DiffractionPatterns,
    SimpleDiffractionDataset,
)
from ptychodus.api.io import AssembledDiffractionData, load_diffraction_data, save_diffraction_data
from ptychodus.api.tree import SimpleTreeNode

from ..task_manager import BackgroundTask, TaskManager
from ._loader import ArrayAssembler, LoadAllArrays, LoadArray
from .bad_pixels import BadPixelsProvider
from .settings import DiffractionSettings
from .sizer import PatternSizer

logger = logging.getLogger(__name__)

__all__ = [
    'AssembledDiffractionArray',
    'AssembledDiffractionDataset',
    'DiffractionDatasetObserver',
]


class DiffractionDatasetObserver(ABC):
    @abstractmethod
    def handle_array_inserted(self, index: int) -> None:
        pass

    @abstractmethod
    def handle_array_changed(self, index: int) -> None:
        pass

    @abstractmethod
    def handle_dataset_reloaded(self) -> None:
        pass


class AssembledDiffractionArray(DiffractionArray):
    def __init__(
        self,
        array_index: int,
        label: str,
        data: AssembledDiffractionData,
    ) -> None:
        super().__init__()
        self._array_index = array_index
        self._label = label
        self._data = data
        self._pattern_counts = data.get_pattern_counts()
        self._average_pattern = data.get_average_pattern()

    @classmethod
    def create_null(cls) -> AssembledDiffractionArray:
        data = AssembledDiffractionData.create_null()
        return cls(0, 'null', data)

    @property
    def array_index(self) -> int:
        return self._array_index

    def get_label(self) -> str:
        return self._label

    def get_indexes(self) -> DiffractionIndexes:
        return self._data.get_indexes()

    def get_patterns(self) -> DiffractionPatterns:
        return self._data.get_patterns()

    def get_pattern(self, index: int) -> DiffractionPattern:
        return self._data.get_pattern(index)

    def get_pattern_counts(self, index: int) -> int:
        return self._pattern_counts[index]

    def get_mean_pattern_counts(self) -> float:
        return numpy.mean(self._pattern_counts).item()

    def get_max_pattern_counts(self) -> int:
        return self._pattern_counts.max().item()

    def get_average_pattern(self) -> DiffractionPattern:
        return self._average_pattern


class AssembledDiffractionDataset(DiffractionDataset, ArrayAssembler):
    def __init__(
        self,
        settings: DiffractionSettings,
        sizer: PatternSizer,
        bad_pixels_provider: BadPixelsProvider,
        task_manager: TaskManager,
    ) -> None:
        super().__init__()
        self._settings = settings
        self._sizer = sizer
        self._bad_pixels_provider = bad_pixels_provider
        self._task_manager = task_manager
        self._observer_list: list[DiffractionDatasetObserver] = []

        self._dataset = SimpleDiffractionDataset.create_null()
        self._data = AssembledDiffractionData.create_null()
        self._array_list: list[AssembledDiffractionArray] = list()
        self._array_counter = 0
        self._array_loader: LoadAllArrays | None = None

    def add_observer(self, observer: DiffractionDatasetObserver) -> None:
        if observer not in self._observer_list:
            self._observer_list.append(observer)

    def remove_observer(self, observer: DiffractionDatasetObserver) -> None:
        try:
            self._observer_list.remove(observer)
        except ValueError:
            pass

    def get_metadata(self) -> DiffractionMetadata:
        return self._dataset.get_metadata()

    def get_layout(self) -> SimpleTreeNode:
        return self._dataset.get_layout()

    def get_bad_pixels(self) -> BadPixels | None:
        return self._dataset.get_bad_pixels()

    def get_assembled_data(self) -> AssembledDiffractionData:
        return self._data

    @overload
    def __getitem__(self, index: int) -> AssembledDiffractionArray: ...

    @overload
    def __getitem__(self, index: slice) -> Sequence[AssembledDiffractionArray]: ...

    def __getitem__(
        self, index: int | slice
    ) -> AssembledDiffractionArray | Sequence[AssembledDiffractionArray]:
        return self._array_list[index]

    def __len__(self) -> int:
        return len(self._array_list)

    def _create_array_loader(
        self, array: DiffractionArray, *, process_patterns: bool
    ) -> BackgroundTask:
        """Load a new array into the dataset. Assumes that arrays arrive in order."""
        bad_pixels = self._dataset.get_bad_pixels()

        if bad_pixels is None:
            raise RuntimeError('Cannot load array without bad pixel map!')

        array_index = self._array_counter
        self._array_counter += 1

        processor = self._sizer.get_processor()
        return LoadArray(
            array_index, array, bad_pixels, processor if process_patterns else None, self
        )

    def append_array(self, array: DiffractionArray, *, process_patterns: bool = True) -> None:
        task = self._create_array_loader(array, process_patterns=process_patterns)
        self._task_manager.put_background_task(task)

    def _insert_array(self, array: AssembledDiffractionArray) -> None:
        pos = bisect(self._array_list, array.array_index, key=lambda x: x.array_index)
        self._array_list.insert(pos, array)

        for observer in self._observer_list:
            observer.handle_array_inserted(pos)

    def _assemble_array(
        self,
        array_index: int,
        label: str,
        data: AssembledDiffractionData,
    ) -> None:
        metadata = self.get_metadata()
        num_patterns_per_array = metadata.num_patterns_per_array
        offset = sum(num_patterns_per_array[:array_index])
        assembled_data_view = self._data.assemble(data, offset)
        assembled_array = AssembledDiffractionArray(
            array_index=array_index,
            label=label,
            data=assembled_data_view,
        )
        self._task_manager.put_foreground_task(lambda: self._insert_array(assembled_array))

    def clear(self) -> None:
        self._dataset = SimpleDiffractionDataset.create_null()
        self._data = AssembledDiffractionData.create_null()
        self._array_list.clear()
        self._array_counter = 0
        self._array_loader = None

        for observer in self._observer_list:
            observer.handle_dataset_reloaded()

    def reload(self, dataset: DiffractionDataset) -> None:
        self.clear()
        self._dataset = SimpleDiffractionDataset(dataset.get_metadata(), dataset.get_layout(), [])
        self._array_loader = LoadAllArrays(dataset, self, self._task_manager)

        for observer in self._observer_list:
            observer.handle_dataset_reloaded()

    def load_all_arrays(self, *, process_patterns: bool, block: bool) -> None:
        if self._array_loader is None:
            logger.warning('Arrays have already been loaded!')
            return

        metadata = self._dataset.get_metadata()

        if metadata.detector_extent is not None:
            self._bad_pixels_provider.set_detector_extent(metadata.detector_extent)

        bad_pixels = self._bad_pixels_provider.get_bad_pixels()

        if process_patterns:
            processor = self._sizer.get_processor()
            bad_pixels = processor.process_bad_pixels(bad_pixels)
            self._array_loader.enable_pattern_processing()

        self._dataset = SimpleDiffractionDataset(
            metadata, self._dataset.get_layout(), [], bad_pixels
        )

        num_patterns_total = sum(metadata.num_patterns_per_array)
        indexes = -numpy.ones(num_patterns_total, dtype=int)

        patterns_shape = num_patterns_total, bad_pixels.shape[-2], bad_pixels.shape[-1]
        patterns_dtype = metadata.pattern_dtype

        if self._settings.memmap_enabled.get_value():
            scratch_dir = self._settings.scratch_directory.get_value()
            scratch_dir.mkdir(mode=0o755, parents=True, exist_ok=True)
            npy_tmp_file = tempfile.NamedTemporaryFile(dir=scratch_dir, suffix='.npy')
            logger.info(f'Scratch data file {npy_tmp_file.name} is {patterns_shape}')
            patterns: DiffractionPatterns = numpy.memmap(
                npy_tmp_file, dtype=patterns_dtype, shape=patterns_shape
            )
            patterns[:] = 0
        else:
            logger.info(f'Scratch memory is {patterns_shape}')
            patterns = numpy.zeros(patterns_shape, dtype=patterns_dtype)
            logger.debug(f'{patterns.nbytes / BYTES_PER_MEGABYTE:.2f}MB allocated for patterns')

        self._data = AssembledDiffractionData(
            indexes=indexes,
            patterns=patterns,
            bad_pixels=bad_pixels,
        )

        # load all arrays in background
        finished_event = self._array_loader.get_finished_event()
        self._task_manager.put_background_task(self._array_loader)
        self._array_loader = None

        if block:
            while not self._task_manager.is_stopping:
                if finished_event.wait(timeout=TaskManager.WAIT_TIME_S):
                    break

    def _generate_dataset_for_assembled_data(self, file_path: Path | None = None) -> None:
        # TODO deconflict detector size with bad_pixels_provider
        num_patterns, detector_height, detector_width = self._data.get_patterns_shape()
        metadata = DiffractionMetadata(
            num_patterns_per_array=[num_patterns],
            pattern_dtype=self._data.get_patterns_dtype(),
            detector_extent=ImageExtent(detector_width, detector_height),
            file_path=file_path,
        )
        contents_tree = SimpleTreeNode.create_root(['Name', 'Type', 'Details'])
        array = AssembledDiffractionArray(
            array_index=0,
            label='In-Memory' if file_path is None else file_path.stem,
            data=self._data,
        )
        bad_pixels = self._data.get_bad_pixels()

        self._dataset = SimpleDiffractionDataset(metadata, contents_tree, [], bad_pixels)
        self._array_list = [array]
        self._array_counter = 1

        for observer in self._observer_list:
            observer.handle_dataset_reloaded()

    def set_assembled_patterns(self, data: AssembledDiffractionData) -> None:
        self.clear()
        self._data = data
        self._generate_dataset_for_assembled_data(file_path=None)

    def import_assembled_patterns(self, file_path: Path) -> None:
        if file_path.is_file():
            self.clear()
            logger.info(f'Importing assembled dataset from "{file_path}"')
            self._data = load_diffraction_data(file_path)
            self._generate_dataset_for_assembled_data(file_path=file_path)
        else:
            logger.warning(f'Refusing to read invalid file path {file_path}')

    def export_assembled_patterns(self, file_path: Path, compression: str = 'lzf') -> None:
        logger.info(f'Exporting assembled dataset to "{file_path}"')
        save_diffraction_data(file_path, self._data, compression=compression)

    def get_info_text(self) -> str:
        file_path = self.get_metadata().file_path
        label = file_path.stem if file_path else 'None'
        return f'{label}: {self._data}'



from collections.abc import Sequence
from pathlib import Path

from PyQt5.QtWidgets import QDialog, QFileDialog, QWidget

from ptychodus.api.observer import Observable


class FileDialogFactory(Observable):
    def __init__(self) -> None:
        super().__init__()
        self._open_working_directory = Path.cwd()
        self._save_working_directory = Path.cwd()

    def get_open_working_directory(self) -> Path:
        return self._open_working_directory

    def set_open_working_directory(self, directory: Path) -> None:
        if not directory.is_dir():
            directory = directory.parent

        directory = directory.resolve()

        if self._open_working_directory != directory:
            self._open_working_directory = directory
            self.notify_observers()

    def get_open_file_path(
        self,
        parent: QWidget,
        caption: str,
        name_filters: Sequence[str] | None = None,
        mime_type_filters: Sequence[str] | None = None,
        selected_name_filter: str | None = None,
    ) -> tuple[Path | None, str]:
        file_path = None

        dialog = QFileDialog(parent, caption, str(self.get_open_working_directory()))
        dialog.setAcceptMode(QFileDialog.AcceptMode.AcceptOpen)
        dialog.setFileMode(QFileDialog.FileMode.ExistingFile)

        if name_filters is not None:
            dialog.setNameFilters(name_filters)

        if mime_type_filters is not None:
            dialog.setMimeTypeFilters(mime_type_filters)

        if selected_name_filter is not None:
            dialog.selectNameFilter(selected_name_filter)

        if dialog.exec() == QDialog.DialogCode.Accepted:  # TODO exec -> open
            file_name_list = dialog.selectedFiles()
            file_name = file_name_list[0]

            if file_name:
                file_path = Path(file_name)
                self.set_open_working_directory(file_path.parent)

        return file_path, dialog.selectedNameFilter()

    def get_save_file_path(
        self,
        parent: QWidget,
        caption: str,
        name_filters: Sequence[str] | None = None,
        mime_type_filters: Sequence[str] | None = None,
        selected_name_filter: str | None = None,
    ) -> tuple[Path | None, str]:
        file_path = None

        dialog = QFileDialog(parent, caption, str(self._save_working_directory))
        dialog.setAcceptMode(QFileDialog.AcceptMode.AcceptSave)
        dialog.setFileMode(QFileDialog.FileMode.AnyFile)

        if name_filters is not None:
            dialog.setNameFilters(name_filters)

        if mime_type_filters is not None:
            dialog.setMimeTypeFilters(mime_type_filters)

        if selected_name_filter is not None:
            dialog.selectNameFilter(selected_name_filter)

        if dialog.exec() == QDialog.DialogCode.Accepted:  # TODO exec -> open
            file_name_list = dialog.selectedFiles()
            file_name = file_name_list[0]

            if file_name:
                file_path = Path(file_name)
                self._save_working_directory = file_path.parent

        return file_path, dialog.selectedNameFilter()

    def get_existing_directory_path(
        self, parent: QWidget, caption: str, initial_directory: Path | None = None
    ) -> Path | None:
        dir_path = None

        dir_name = QFileDialog.getExistingDirectory(
            parent,
            caption,
            str(initial_directory or self.get_open_working_directory()),
            QFileDialog.Option.ShowDirsOnly | QFileDialog.Option.DontResolveSymlinks,
        )

        if dir_name:
            dir_path = Path(dir_name)
            self.set_open_working_directory(dir_path)

        return dir_path



# mypy: ignore-errors

import numpy
from numpy.typing import NDArray
from scipy import signal, ndimage
from typing import Literal, Optional, Tuple


class PhaseUnwrapper:
    def __init__(
        self,
        fourier_shift_step: float = 0.5,
        image_grad_method: Literal[
            'fourier_shift', 'fourier_differentiation', 'nearest'
        ] = 'fourier_differentiation',
        image_integration_method: Literal['fourier', 'discrete', 'deconvolution'] = 'fourier',
        weight_map: Optional[NDArray] = None,
        eps: float = 1e-9,
    ) -> None:
        """Get the unwrapped phase of a complex 2D image.

        Parameters
        ----------
        fourier_shift_step : float
            The finite-difference step size used to calculate the gradient,
            if the Fourier shift method is used.
        image_grad_method : str
            The method used to calculate the phase gradient.
                - "fourier_shift": Use Fourier shift to perform shift.
                - "nearest": Use nearest neighbor to perform shift.
                - "fourier_differentiation": Use Fourier differentiation.
        image_integration_method : str
            The method used to integrate the image back from gradients.
                - "fourier": Use Fourier integration as implemented in PtychoShelves.
                - "deconvolution": Deconvolve ramp filter.
                - "discrete": Use cumulative sum.
        weight_map : Optional[NDArray]
            A weight map multiplied to the input image.
        eps : float
            A small number to avoid division by zero.
        """
        self.fourier_shift_step = fourier_shift_step
        self.image_grad_method = image_grad_method
        self.image_integration_method = image_integration_method
        self.weight_map = weight_map
        self.eps = eps

    def unwrap(self, img: NDArray) -> NDArray:
        """Run unwrapping.

        Parameters
        ----------
        img : NDArray
            A 2D complex array giving the image to be unwrapped.

        Returns
        -------
        NDArray
            A 2D real array giving the unwrapped phase of the input image.
        """
        if not numpy.iscomplexobj(img):
            raise ValueError('Input array must be complex.')

        if self.weight_map is not None:
            weight_map = float(numpy.clip(self.weight_map, 0.0, 1.0))
        else:
            weight_map = 1.0

        img = weight_map * img / (numpy.abs(img) + self.eps)
        bc_center = numpy.angle(img[img.shape[0] // 2, img.shape[1] // 2])

        # Pad image to avoid FFT boundary artifacts.
        padding = [64, 64]
        if any(numpy.array(padding) > 0):
            img = numpy.pad(
                img, ((padding[0], padding[0]), (padding[1], padding[1])), mode='reflect'
            )
            img = vignett(img, margin=10, sigma=2.5)

        gy, gx = get_phase_gradient(
            img,
            fourier_shift_step=self.fourier_shift_step,
            image_grad_method=self.image_grad_method,
        )

        if self.image_integration_method == 'discrete' and any(numpy.array(padding) > 0):
            gy = gy[padding[0] : -padding[0], padding[1] : -padding[1]]
            gx = gx[padding[0] : -padding[0], padding[1] : -padding[1]]
        if self.image_integration_method == 'discrete':
            phase = numpy.real(integrate_image_2d(gy, gx, bc_center=bc_center))
        elif self.image_integration_method == 'fourier':
            phase = numpy.real(integrate_image_2d_fourier(gy, gx))
        elif self.image_integration_method == 'deconvolution':
            phase = numpy.real(integrate_image_2d_deconvolution(gy, gx, bc_center=bc_center))
        else:
            raise ValueError(f'Unknown integration method: {self.image_integration_method}')

        if self.image_integration_method != 'discrete' and any(numpy.array(padding) > 0):
            gy = gy[padding[0] : -padding[0], padding[1] : -padding[1]]
            gx = gx[padding[0] : -padding[0], padding[1] : -padding[1]]
            phase = phase[padding[0] : -padding[0], padding[1] : -padding[1]]

        return phase


def vignett(img: NDArray, margin: int = 20, sigma: float = 1.0) -> NDArray:
    """Vignett an image so that it gradually decays near the boundary.
    For each dimension of the image, a mask with a width of `2 * margin`
    and with half of it filled with 0s and half with 1s is
    generated and convolved with a Gaussian kernel of size
    `margin` and standard deviation `sigma`. The blurred mask is cropped and
    multiplied to the near-edge regions of the image.

    Parameters
    ----------
    img : Tensor
        The input image.
    margin : int
        The margin of image where the decay takes place.
    sigma : float
        The standard deviation of the Gaussian kernel.
    """
    img = img.copy()
    for i_dim in range(img.ndim):
        if img.shape[i_dim] <= 2 * margin:
            continue

        mask_shape = (
            [img.shape[i] for i in range(i_dim)]
            + [2 * margin]
            + [img.shape[i] for i in range(i_dim + 1, img.ndim)]
        )
        mask = numpy.zeros(mask_shape)
        mask_slicer = [slice(None)] * i_dim + [slice(margin, None)]
        mask[tuple(mask_slicer)] = 1.0

        gauss_win = signal.windows.gaussian(margin // 2, std=sigma)
        gauss_win = gauss_win / numpy.sum(gauss_win)
        mask = ndimage.convolve1d(mask, gauss_win, axis=i_dim, mode='constant')
        mask_final_slicer = [slice(None)] * i_dim + [slice(len(gauss_win), len(gauss_win) + margin)]

        mask = mask[tuple(mask_final_slicer)]

        mask = numpy.where(mask < 1e-3, 0, mask)

        slicer = tuple([slice(None)] * i_dim + [slice(0, margin)])
        img[slicer] = img[slicer] * mask

        slicer = tuple([slice(None)] * i_dim + [slice(-margin, None)])
        img[slicer] = img[slicer] * numpy.flip(mask, axis=i_dim)
    return img


def nearest_neighbor_gradient(
    image: NDArray, direction: Literal['forward', 'backward'], dim: Tuple[int, ...] = (0, 1)
) -> Tuple[NDArray, NDArray]:
    """
    Calculate the nearest neighbor gradient of a 2D image.

    Parameters
    ----------
    image : NDArray
        a (... H, W) tensor of images.
    direction : str
        'forward' or 'backward'.
    dim : tuple of int, optional
        Dimensions to calculate gradient. Default is (0, 1).

    Returns
    -------
    tuple of NDArray
        a tuple of 2 images with the gradient in y and x directions.
    """
    if not hasattr(dim, '__len__'):
        dim = (dim,)
    grad_x = None
    grad_y = None
    if direction == 'forward':
        if 1 in dim:
            grad_x = numpy.concatenate([image[:, 1:], image[:, -1:]], axis=1) - image
        if 0 in dim:
            grad_y = numpy.concatenate([image[1:, :], image[-1:, :]], axis=0) - image
    elif direction == 'backward':
        if 1 in dim:
            grad_x = image - numpy.concatenate([image[:, :1], image[:, :-1]], axis=1)
        if 0 in dim:
            grad_y = image - numpy.concatenate([image[:1, :], image[:-1, :]], axis=0)
    else:
        raise ValueError("direction must be 'forward' or 'backward'")
    return grad_y, grad_x


def gaussian_gradient(image: NDArray, sigma: float = 1.0, kernel_size=5) -> Tuple[NDArray, NDArray]:
    """
    Calculate the gradient of a 2D image with a Gaussian-derivative kernel.

    Parameters
    ----------
    image : NDArray
        A (... H, W) tensor of images.
    sigma : float
        Sigma of the Gaussian.

    Returns
    -------
    tuple of NDArray
        A tuple of 2 images with the gradient in y and x directions.
    """
    r = numpy.arange(kernel_size) - (kernel_size - 1) / 2.0
    kernel = -r / (numpy.sqrt(2 * numpy.pi) * sigma**3) * numpy.exp(-(r**2) / (2 * sigma**2))
    grad_y = ndimage.convolve(image, kernel.reshape(-1, 1), mode='nearest')
    grad_x = ndimage.convolve(image, kernel.reshape(1, -1), mode='nearest')

    # Gate the gradients
    grads = [grad_y, grad_x]
    for i, g in enumerate(grads):
        m = numpy.logical_and(numpy.abs(grad_y) < 1e-6, numpy.abs(grad_y) != 0)
        if numpy.count_nonzero(m) > 0:
            print('Gradient magnitudes between 0 and 1e-6 are set to 0.')
            g = g * numpy.logical_not(m)
            grads[i] = g
    grad_y, grad_x = grads
    return grad_y, grad_x


def fourier_gradient(image: NDArray) -> Tuple[NDArray, NDArray]:
    """Calculate gradient using NumPy FFT operations"""
    u = numpy.fft.fftfreq(image.shape[0])
    v = numpy.fft.fftfreq(image.shape[1])
    u, v = numpy.meshgrid(u, v, indexing='ij')

    grad_y = numpy.fft.ifft(numpy.fft.fft(image, axis=-2) * (2j * numpy.pi * u), axis=-2)
    grad_x = numpy.fft.ifft(numpy.fft.fft(image, axis=-1) * (2j * numpy.pi * v), axis=-1)

    return grad_y, grad_x


def get_phase_gradient(
    img: NDArray,
    fourier_shift_step: float = 0,
    image_grad_method: Literal[
        'fourier_shift', 'fourier_differentiation', 'nearest'
    ] = 'fourier_shift',
    eps: float = 1e-6,
) -> Tuple[NDArray, NDArray]:
    """
    Get the gradient of the phase of a complex 2D image by first calculating
    the spatial gradient of the complex image, then taking the phase of the
    complex gradient -- i.e., it takes the phase of the gradient rather than
    the gradient of the phase. This avoids the sharp gradients due to phase
    wrapping when directly taking the gradient of the phase.

    Parameters
    ----------
    img : NDArray
        A [N, H, W] or [H, W] tensor giving a batch of images or a single image.
    step : float
        The finite-difference step size used to calculate the gradient, if
        the Fourier shift method is used.
    finite_diff_method : enums.ImageGradientMethods
        The method used to calculate the phase gradient.
            - "fourier_shift": Use Fourier shift to perform shift.
            - "nearest": Use nearest neighbor to perform shift.
            - "fourier_differentiation": Use Fourier differentiation.
    eps : float
        A stablizing constant.

    Returns
    -------
    Tuple[NDArray, NDArray]
        A tuple of 2 images with the gradient in y and x directions.
    """
    if fourier_shift_step <= 0 and image_grad_method == 'fourier_shift':
        raise ValueError('Step must be positive.')

    if image_grad_method == 'fourier_differentiation':
        gy, gx = fourier_gradient(img)
        gy = numpy.imag(numpy.conj(img) * gy)
        gx = numpy.imag(numpy.conj(img) * gx)
    else:
        # Use finite difference.
        if img.ndim == 2:
            img = img[None, ...]
        pad = int(numpy.ceil(fourier_shift_step)) + 1
        img = numpy.pad(img, ((0, 0), (pad, pad), (pad, pad)), mode='reflect')

        sy1 = numpy.array([[-fourier_shift_step, 0]]).repeat(img.shape[0], axis=0)
        sy2 = numpy.array([[fourier_shift_step, 0]]).repeat(img.shape[0], axis=0)
        if image_grad_method == 'fourier_shift':
            # If the image contains zero-valued pixels, Fourier shift can result in small
            # non-zero values that dangles around 0. This can cause the phase
            # of the shifted image to dangle between pi and -pi. In that case, use
            # `finite_diff_method="nearest" instead`, or use `step=1`.
            complex_prod = fourier_shift(img, sy1) * fourier_shift(img, sy2).conj()
        elif image_grad_method == 'nearest':
            complex_prod = img * numpy.concatenate([img[:, :1, :], img[:, :-1, :]], axis=1).conj()
        else:
            raise ValueError(f'Unknown finite-difference method: {image_grad_method}')
        complex_prod = numpy.where(
            numpy.abs(complex_prod) < numpy.abs(complex_prod).max() * 1e-6, 0, complex_prod
        )
        gy = numpy.angle(complex_prod) / (2 * fourier_shift_step)
        gy = gy[0, pad:-pad, pad:-pad]

        sx1 = numpy.array([[0, -fourier_shift_step]]).repeat(img.shape[0], axis=0)
        sx2 = numpy.array([[0, fourier_shift_step]]).repeat(img.shape[0], axis=0)
        if image_grad_method == 'fourier_shift':
            complex_prod = fourier_shift(img, sx1) * fourier_shift(img, sx2).conj()
        elif image_grad_method == 'nearest':
            complex_prod = img * numpy.concatenate([img[:, :, :1], img[:, :, :-1]], axis=2).conj()
        complex_prod = numpy.where(
            numpy.abs(complex_prod) < numpy.abs(complex_prod).max() * 1e-6, 0, complex_prod
        )
        gx = numpy.angle(complex_prod) / (2 * fourier_shift_step)
        gx = gx[0, pad:-pad, pad:-pad]
    return gy, gx


def integrate_image_2d_fourier(grad_y: NDArray, grad_x: NDArray) -> NDArray:
    """
    Integrate an image with the gradient in y and x directions using Fourier
    differentiation.

    Parameters
    ----------
    grad_y, grad_x: NDArray
        A (H, W) tensor of gradients in y or x directions.

    Returns
    -------
    NDArray
        The integrated image.
    """
    shape = grad_y.shape
    f = numpy.fft.fft2(grad_x + 1j * grad_y)
    y, x = numpy.fft.fftfreq(shape[0]), numpy.fft.fftfreq(shape[1])

    r = 1.0
    r = r / (2j * numpy.pi * (x + 1j * y[:, None]) + 1e-15)
    r[0, 0] = 0
    integrated_image = f * r
    integrated_image = numpy.fft.ifft2(integrated_image)
    if not numpy.iscomplexobj(grad_x):
        integrated_image = integrated_image.real
    return integrated_image


def integrate_image_2d_deconvolution(
    grad_y: NDArray,
    grad_x: NDArray,
    tf_y: Optional[NDArray] = None,
    tf_x: Optional[NDArray] = None,
    bc_center: float = 0,
) -> NDArray:
    """
    Integrate an image with the gradient in y and x directions by deconvolving
    the differentiation kernel, whose transfer function is assumed to be a
    ramp function.

    Adapted from Tripathi, A., McNulty, I., Munson, T., & Wild, S. M. (2016).
    Single-view phase retrieval of an extended sample by exploiting edge detection
    and sparsity. Optics Express, 24(21), 24719–24738. doi:10.1364/OE.24.024719

    Parameters
    ----------
    grad_y, grad_x: NDArray
        A (H, W) tensor of gradients in y or x directions.
    tf_y, tf_x: NDArray
        A (H, W) tensor of transfer functions in y or x directions. If not
        provided, they are assumed to be 2i * pi * u (or v), which are the
        effective transfer functions in Fourier differentiation.
    bc_center: float
        The value of the boundary condition at the center of the image.

    Returns
    -------
    NDArray
        The integrated image.
    """
    u, v = numpy.fft.fftfreq(grad_x.shape[0]), numpy.fft.fftfreq(grad_x.shape[1])
    u, v = numpy.meshgrid(u, v, indexing='ij')
    if tf_y is None or tf_x is None:
        tf_y = 2j * numpy.pi * u
        tf_x = 2j * numpy.pi * v
    f_grad_y = numpy.fft.fft2(grad_y)
    f_grad_x = numpy.fft.fft2(grad_x)
    img = (f_grad_y * tf_y + f_grad_x * tf_x) / (numpy.abs(tf_y) ** 2 + numpy.abs(tf_x) ** 2 + 1e-5)
    img = -numpy.fft.ifft2(img)
    img = img + bc_center - img[img.shape[0] // 2, img.shape[1] // 2]
    return img


def integrate_image_2d(grad_y: NDArray, grad_x: NDArray, bc_center: float = 0) -> NDArray:
    """
    Integrate an image with the gradient in y and x directions.

    Parameters
    ----------
    grad_y : NDArray
        The gradient in y direction.
    grad_x : NDArray
        The gradient in x direction.
    bc_center : float
        The boundary condition at the center of the image, by default 0

    Returns
    -------
    NDArray
        The integrated image.
    """
    left_boundary = numpy.cumsum(grad_y[:, 0], axis=0)
    int_img = numpy.cumsum(grad_x, axis=1) + left_boundary[:, None]
    int_img = int_img + bc_center - int_img[int_img.shape[0] // 2, int_img.shape[1] // 2]
    return int_img


def fourier_shift(
    images: NDArray, shifts: NDArray, strictly_preserve_zeros: bool = False
) -> NDArray:
    """
    Apply Fourier shift to a batch of images.

    Parameters
    ----------
    images : NDArray
        A [N, H, W] array of images.
    shifts : NDArray
        A [N, 2] array of shifts in pixels.
    strictly_preserve_zeros : bool
        If True, mask of strictly zero pixels will be generated and shifted
        by the same amount. Pixels that have a non-zero value in the shifted
        mask will be set to zero in the shifted image. This preserves the zero
        pixels in the original image, preventing FFT from introducing small
        non-zero values due to machine precision.

    Returns
    -------
    NDArray
        Shifted images.
    """
    if strictly_preserve_zeros:
        zero_mask = images == 0
        zero_mask = zero_mask.float()
        zero_mask_shifted = fourier_shift(zero_mask, shifts, strictly_preserve_zeros=False)
    ft_images = numpy.fft.fft2(images)
    freq_y, freq_x = numpy.meshgrid(
        numpy.fft.fftfreq(images.shape[-2]), numpy.fft.fftfreq(images.shape[-1]), indexing='ij'
    )
    freq_x = freq_x.repeat(images.shape[0], axis=0)
    freq_y = freq_y.repeat(images.shape[0], axis=0)
    mult = numpy.exp(
        1j
        * -2
        * numpy.pi
        * (freq_x * shifts[:, 1].reshape([-1, 1, 1]) + freq_y * shifts[:, 0].reshape([-1, 1, 1]))
    )
    ft_images = ft_images * mult
    shifted_images = numpy.fft.ifft2(ft_images)
    if not numpy.iscomplexobj(images):
        shifted_images = shifted_images.real
    if strictly_preserve_zeros:
        shifted_images[zero_mask_shifted > 0] = 0
    return shifted_images



from __future__ import annotations
from pathlib import Path
import logging

import watchdog.events
from watchdog.observers.polling import PollingObserver
import watchdog.observers

from ptychodus.api.observer import Observable, Observer

from .buffer import PendingFileBuffer
from .settings import AutomationSettings

logger = logging.getLogger(__name__)


class DataDirectoryEventHandler(watchdog.events.FileSystemEventHandler):
    def __init__(self, dataset_buffer: PendingFileBuffer) -> None:
        super().__init__()
        self._dataset_buffer = dataset_buffer

    def on_created_or_modified(self, event: watchdog.events.FileSystemEvent) -> None:
        if not event.is_directory:
            src_path = Path(str(event.src_path))
            current_workflow = self._dataset_buffer.get_current_workflow()

            if src_path.match(current_workflow.get_watch_file_pattern()):
                self._dataset_buffer.put(src_path)

    def on_created(self, event: watchdog.events.FileSystemEvent) -> None:
        self.on_created_or_modified(event)

    def on_modified(self, event: watchdog.events.FileSystemEvent) -> None:
        self.on_created_or_modified(event)


class DataDirectoryWatcher(Observable, Observer):
    def __init__(
        self,
        settings: AutomationSettings,
        dataset_buffer: PendingFileBuffer,
    ) -> None:
        super().__init__()
        self._settings = settings
        self._dataset_buffer = dataset_buffer
        self._observer = watchdog.observers.Observer()

        settings.add_observer(self)
        dataset_buffer.add_observer(self)

    @property
    def is_alive(self) -> bool:
        return self._observer.is_alive()

    def _update_watch(self) -> None:
        self._observer.unschedule_all()
        data_directory = self._settings.data_directory.get_value()

        if data_directory.exists():
            current_workflow = self._dataset_buffer.get_current_workflow()
            observed_watch = self._observer.schedule(
                event_handler=DataDirectoryEventHandler(self._dataset_buffer),
                path=str(data_directory),
                recursive=current_workflow.is_watch_recursive,
            )
            logger.debug(observed_watch)
        else:
            logger.warning(f'Data directory "{data_directory}" does not exist!')

    def start(self) -> None:
        if self.is_alive:
            logger.error('Automation watchdog thread already started!')
        else:
            logger.debug('Starting automation watchdog thread...')
            self._observer = (
                PollingObserver()
                if self._settings.use_watchdog_polling_observer.get_value()
                else watchdog.observers.Observer()
            )
            self._observer.start()
            self._update_watch()
            logger.debug('Automation watchdog thread started.')

    def stop(self) -> None:
        if self.is_alive:
            logger.debug('Stopping automation watchdog thread...')
            self._observer.stop()
            self._observer.join()
            logger.debug('Automation watchdog thread stopped.')
        else:
            logger.debug('Automation watchdog already stopped.')

    def _update(self, observable: Observable) -> None:
        if observable is self._settings:
            self._update_watch()
        elif observable is self._dataset_buffer:
            self._update_watch()



from __future__ import annotations
import logging

import numpy

from PyQt5.QtCore import Qt, QLineF, QRectF
from PyQt5.QtWidgets import QGraphicsScene, QStatusBar

from ptychodus.api.common import NumberArrayType
from ptychodus.api.geometry import Box2D, Line2D, PixelGeometry, Point2D
from ptychodus.api.observer import Observable, Observer

from ...model.visualization import VisualizationEngine
from ...view.visualization import (
    HistogramDialog,
    ImageItem,
    ImageItemSignals,
    ImageMouseTool,
    LineCutDialog,
    VisualizationView,
)
from ...view.widgets import ExceptionDialog
from ..data import FileDialogFactory

logger = logging.getLogger(__name__)


class VisualizationController(Observer):
    MIME_TYPES = ['image/bmp', 'image/jpeg', 'image/png', 'image/x-portable-pixmap']

    def __init__(
        self,
        engine: VisualizationEngine,
        view: VisualizationView,
        status_bar: QStatusBar,
        file_dialog_factory: FileDialogFactory,
    ) -> None:
        super().__init__()
        self._engine = engine
        self._view = view
        self._status_bar = status_bar
        self._file_dialog_factory = file_dialog_factory
        self._line_cut_dialog = LineCutDialog(view)
        self._histogram_dialog = HistogramDialog(view)

        item_signals = ImageItemSignals()
        item_signals.line_cut_finished.connect(self._analyze_line_cut)
        item_signals.rectangle_finished.connect(self._analyze_region)

        self._item = ImageItem(item_signals, status_bar)
        engine.add_observer(self)

        scene = QGraphicsScene()
        scene.addItem(self._item)
        view.setScene(scene)

        view.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        view.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)

    def get_item(self) -> ImageItem:
        return self._item

    def set_array(
        self,
        array: NumberArrayType,
        pixel_geometry: PixelGeometry,
        *,
        autoscale_color_axis: bool = False,
    ) -> None:
        if numpy.all(numpy.isfinite(array)):
            try:
                product = self._engine.render(
                    array, pixel_geometry, autoscale_color_axis=autoscale_color_axis
                )
            except ValueError as err:
                logger.exception(err)
                ExceptionDialog.show_exception('Renderer', err)
            else:
                self._item.set_product(product)
        else:
            logger.warning('Array contains infinite or NaN values!')
            self._item.clear_product()

    def clear_array(self) -> None:
        self._item.clear_product()

    def set_mouse_tool(self, mouse_tool: ImageMouseTool) -> None:
        self._item.set_mouse_tool(mouse_tool)

    def save_image(self) -> None:
        file_path, _ = self._file_dialog_factory.get_save_file_path(
            self._view, 'Save Image', mime_type_filters=VisualizationController.MIME_TYPES
        )

        if file_path:
            pixmap = self._item.pixmap()
            pixmap.save(str(file_path))

    def _analyze_line_cut(self, line: QLineF) -> None:
        p1 = Point2D(line.x1(), line.y1())
        p2 = Point2D(line.x2(), line.y2())
        line2d = Line2D(p1, p2)

        product = self._item.get_product()

        if product is None:
            logger.warning('No visualization product!')
            return

        value_label = product.get_value_label()
        line_cut = product.get_line_cut(line2d)

        ax = self._line_cut_dialog.axes
        ax.clear()
        ax.plot(line_cut.distance_m, line_cut.value, '.-', linewidth=1.5)
        ax.set_xlabel('Distance [m]')
        ax.set_ylabel(value_label)
        ax.grid(True)
        self._line_cut_dialog.figure_canvas.draw()
        self._line_cut_dialog.open()

    def _analyze_region(self, rect: QRectF) -> None:
        if rect.isEmpty():
            logger.debug('QRectF is empty!')
            return

        box = Box2D(
            x=rect.x(),
            y=rect.y(),
            width=rect.width(),
            height=rect.height(),
        )

        product = self._item.get_product()

        if product is None:
            logger.warning('No visualization product!')
            return

        value_label = product.get_value_label()
        kde = product.estimate_kernel_density(box)
        values = numpy.linspace(kde.value_lower, kde.value_upper, 1000)

        ax = self._histogram_dialog.axes
        ax.clear()
        ax.plot(values, kde.kde(values), '.-', linewidth=1.5)
        ax.set_xlabel(value_label)
        ax.set_ylabel('Density')
        ax.grid(True)
        self._histogram_dialog.figure_canvas.draw()

        rectangle_view = self._histogram_dialog.rectangle_view
        rect_center = rect.center()
        rectangle_view.center_x_line_edit.setText(f'{rect_center.x():.1f}')
        rectangle_view.center_y_line_edit.setText(f'{rect_center.y():.1f}')
        rectangle_view.width_line_edit.setText(f'{rect.width():.1f}')
        rectangle_view.height_line_edit.setText(f'{rect.height():.1f}')

        # TODO use rect for crop
        self._histogram_dialog.open()

    def zoom_to_fit(self) -> None:
        self._item.setPos(0, 0)
        scene = self._view.scene()

        if scene is None:
            raise ValueError('scene is None!')

        bounding_rect = scene.itemsBoundingRect()
        scene.setSceneRect(bounding_rect)
        self._view.fitInView(scene.sceneRect(), Qt.AspectRatioMode.KeepAspectRatio)

    def rerender_image(self, *, autoscale_color_axis: bool = False) -> None:
        product = self._item.get_product()

        if product is not None:
            self.set_array(
                product.get_values(),
                product.get_pixel_geometry(),
                autoscale_color_axis=autoscale_color_axis,
            )

    def _update(self, observable: Observable) -> None:
        if observable is self._engine:
            self.rerender_image()
